#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko flexibility typer
#

from copy import deepcopy
from collections import defaultdict
from itertools import product
from operator import itemgetter
from .utils import pdbutils


class FlexibilityBuilder:

    def __call__(self, setup, freeze_bonds=None, root_atom_index=None, break_combo_data=None, bonds_in_rigid_rings=None, glue_pseudo_atoms=None):
        """ """
        self.setup = setup
        self.flexibility_models = {}
        self._frozen_bonds = []
        if not freeze_bonds is None:
            self._frozen_bonds = freeze_bonds[:]
        # build graph for standard molecule (no open macrocycle rings)
        model = self.build_rigid_body_connectivity()
        model = self.set_graph_root(model, root_atom_index) # finds root if root_atom_index==None
        self.add_flex_model(model, score=False)

        # evaluate possible graphs for various ring openings
        if break_combo_data is not None:
            bond_break_combos = break_combo_data['bond_break_combos']
            bond_break_scores = break_combo_data['bond_break_scores']
            broken_rings_list = break_combo_data['broken_rings']
            for index in range(len(bond_break_combos)):
                bond_break_combo = bond_break_combos[index]
                bond_break_score = bond_break_scores[index]
                broken_rings =     broken_rings_list[index]
                model = self.build_rigid_body_connectivity(bond_break_combo, broken_rings, bonds_in_rigid_rings, glue_pseudo_atoms)
                self.set_graph_root(model, root_atom_index) # finds root if root_atom_index==None
                self.add_flex_model(model, score=True, initial_score=bond_break_score)

        self.select_best_model()

        # clean up
        del self._frozen_bonds
        del self.flexibility_models
        return self.setup


    def select_best_model(self):
        """
        select flexibility model with best complexity score
        """
        if len(self.flexibility_models) == 1: # no macrocyle open rings
            best_model = list(self.flexibility_models.values())[0]
        else:
            score_sorted_models = []
            for m_id, model in list(self.flexibility_models.items()):
                score_sorted_models.append((m_id, model['score']))
            #print("SORTED", score_sorted_models)
            score_sorted = sorted(score_sorted_models, key=itemgetter(1), reverse=True)
            #for model_id, score in score_sorted:
            #    print("ModelId[% 3d] score: %2.2f" % (model_id, score))
            # the 0-model is the rigid model used as reference
            best_model_id, best_model_score = score_sorted[1]
            best_model = self.flexibility_models[best_model_id]
        
        setup = best_model['setup']
        del best_model['setup']
        self.setup = setup
        best_model['torsions_org'] = self.flexibility_models[0]['torsions']

        self.setup.flexibility_model = best_model

    def add_flex_model(self, model, score=False, initial_score=0):
        """ add a flexible model to the list of configurations,
            and optionally score it, basing on the connectivity properties
        """

        model_id = len(self.flexibility_models)
        if score == False:
            model['score'] = float('inf')
        else:
            penalty = self.score_flex_model(model)
            model['score'] = initial_score + penalty
        self.flexibility_models[model_id] = model

    def build_rigid_body_connectivity(self, bonds_to_break=None, broken_rings=None, bonds_in_rigid_rings=None, glue_pseudo_atoms=None):
        """
        rigid_body_graph is the graph of rigid bodies
        ( rigid_body_id->[rigid_body_id,...] )

        rigid_body_members contains the atom indices in each rigid_body_id,
        ( rigid_body_id->[atom,...] )

        rigid_body_connectivity contains connectivity information between
        rigid bodies, mapping a two rigid bodies to the two atoms that connect
        them
        ( (rigid_body1, rigid_body2) -> (atom1,atom2)
        """
        # make a copy of the current mol graph, updated with the broken bond
        if bonds_to_break is None:
            self._current_setup = self.setup
        else:
            self._current_setup = self.copy_setup(bonds_to_break, broken_rings, bonds_in_rigid_rings)
            self.update_closure_atoms(bonds_to_break, glue_pseudo_atoms)

        # walk the mol graph to build the rigid body maps
        self._visited = defaultdict(lambda:False)
        self._rigid_body_members = {}
        self._rigid_body_connectivity = {}
        self._rigid_body_graph = defaultdict(list)
        self._rigid_index_by_atom = {}
        # START VALUE HERE SHOULD BE MADE MODIFIABLE FOR FLEX CHAIN
        self._rigid_body_count = 0
        self.walk_rigid_body_graph(start=0)
        # if only a rigid body is found
        if len(self._rigid_body_members) == 1:
            self._rigid_body_connectivity[0] = [0]
            self._rigid_body_graph[0] = [0]
        model = {'rigid_body_graph' : deepcopy(self._rigid_body_graph),
                'rigid_body_connectivity' : deepcopy(self._rigid_body_connectivity),
                'rigid_body_members' : deepcopy(self._rigid_body_members),
                'setup' : self._current_setup}
        return model

    def copy_setup(self, bond_list, broken_rings, bonds_in_rigid_rings):
        """ copy connectivity information (graph and bonds) from the setup,
            optionally delete bond_id listed in bonds_to_break,
            updating connectivty information
        """
        setup = self.setup.copy()
        for bond in bond_list:
            setup.del_bond(*bond)

        for ring in broken_rings:
            for bond in setup.get_bonds_in_ring(ring):
                if bond not in bonds_in_rigid_rings: # e.g. bonds in small rings do not rotata
                    if bond in bond_list:
                        continue # this bond has been deleted
                    bond_item = setup.get_bond(*bond)
                    if bond_item['bond_order'] == 1:
                        setup.bond[bond]['rotatable'] = True
        return setup


    def calc_max_depth(self, graph, seed_node, visited=[], depth=0):
        maxdepth = depth 
        visited.append(seed_node)
        for node in graph[seed_node]:
            if node not in visited:
                visited.append(node)
                newdepth = self.calc_max_depth(graph, node, visited, depth + 1)
                maxdepth = max(maxdepth, newdepth)
        return maxdepth


    def set_graph_root(self, model, root_atom_index=None):
        """ TODO this has to be made aware of the weight of the groups left
         (see 1jff:TA1)
        """

        if root_atom_index is None: # find rigid group that minimizes max_depth
            graph = deepcopy(model['rigid_body_graph'])
            while len(graph) > 2: # remove leafs until 1 or 2 rigid groups remain
                leaves = []
                for vertex, edges in list(graph.items()):
                    if len(edges) == 1:
                        leaves.append(vertex)
                for l in leaves:
                    for vertex, edges in list(graph.items()):
                        if l in edges:
                            edges.remove(l)
                            graph[vertex] = edges
                    del graph[l]

            if len(graph) == 1:
                root_body_index = list(graph.keys())[0]
            else:
                r1, r2 = list(graph.keys())
                r1_size = len(model['rigid_body_members'][r1])
                r2_size = len(model['rigid_body_members'][r2])
                if r1_size >= r2_size:
                    root_body_index = r1
                else:
                    root_body_index = r2

        else: # find index of rigid group
            for body_index in model['rigid_body_members']:
                if root_atom_index in model['rigid_body_members'][body_index]: # 1-index atoms
                    root_body_index = body_index

        model['root'] = root_body_index
        model['torsions'] = len(model['rigid_body_members']) - 1
        model['graph_depth'] = self.calc_max_depth(model['rigid_body_graph'], root_body_index, visited=[], depth=0)

        return model


    def score_flex_model(self, model):
        """ score a flexibility model basing on the graph properties"""
        base = self.flexibility_models[0]['graph_depth']
        score = 10 * (base-model['graph_depth'])
        return score


    def _generate_closure_pseudo(self, setup, bond_id, coords_dict={}):
        """ calculate position and parameters of the pseudoatoms for the closure"""
        closure_pseudo = []
        for idx in (0,1):
            target = bond_id[1 - idx]
            anchor = bond_id[0 - idx]
            if coords_dict is None or len(coords_dict) == 0:
                coord = setup.get_coord(target)
            else:
                coord = coords_dict[anchor]
            anchor_info = setup.pdbinfo[anchor]
            pdbinfo = pdbutils.PDBAtomInfo('G', anchor_info.resName, anchor_info.resNum, anchor_info.chain)
            closure_pseudo.append({
                'coord': coord,
                'anchor_list': [anchor],
                'charge': 0.0,
                'pdbinfo': pdbinfo,
                'atom_type': 'G',
                'bond_type': 0,
                'rotatable': False})
        return closure_pseudo


    def update_closure_atoms(self, bonds_to_break, coords_dict):
        """ create pseudoatoms required by the flexible model with broken bonds"""

        setup = self._current_setup
        for i, bond in enumerate(bonds_to_break):
            setup.ring_closure_info["bonds_removed"].append(bond) # bond is pair of atom indices
            pseudos = self._generate_closure_pseudo(setup, bond, coords_dict)
            for pseudo in pseudos:
                pseudo['atom_type'] = "%s%d" % (pseudo['atom_type'], i)
                pseudo_index = setup.add_pseudo(**pseudo)
                atom_index = pseudo['anchor_list'][0]
                if atom_index in setup.ring_closure_info:
                    raise RuntimeError("did not expect more than one G per atom")
                setup.ring_closure_info["pseudos_by_atom"][atom_index] = pseudo_index
            setup.set_atom_type(bond[0], "CG%d" % i)
            setup.set_atom_type(bond[1], "CG%d" % i)


    def walk_rigid_body_graph(self, start):
        """ recursive walk to build the graph of rigid bodies"""
        idx = 0
        rigid = [start]
        self._visited[start] = True
        current_rigid_body_count = self._rigid_body_count
        self._rigid_index_by_atom[start] = current_rigid_body_count
        sprouts_buffer = []
        while idx < len(rigid):
            current = rigid[idx]
            for neigh in self._current_setup.get_neigh(current):
                bond_id = self._current_setup.get_bond_id(current, neigh)
                bond_info = self._current_setup.get_bond(current, neigh)
                if self._visited[neigh]:
                    is_rigid_bond = (bond_info['rotatable'] == False) or (bond_id in self._frozen_bonds)
                    neigh_in_other_rigid_body = current_rigid_body_count != self._rigid_index_by_atom[neigh]
                    if is_rigid_bond and neigh_in_other_rigid_body:
                        raise RuntimeError('Flexible bonds within rigid group. We have a problem.')
                    continue
                if bond_info['rotatable'] and (bond_id not in self._frozen_bonds):
                    sprouts_buffer.append((current, neigh))
                else:
                    rigid.append(neigh)
                    self._rigid_index_by_atom[neigh] = current_rigid_body_count
                    self._visited[neigh] = True
            idx += 1
        self._rigid_body_members[current_rigid_body_count] = rigid
        for current, neigh in sprouts_buffer:
            if self._visited[neigh]: continue
            self._rigid_body_count+=1
            self._rigid_body_connectivity[current_rigid_body_count, self._rigid_body_count] = current, neigh
            self._rigid_body_connectivity[self._rigid_body_count, current_rigid_body_count] = neigh, current
            self._rigid_body_graph[current_rigid_body_count].append(self._rigid_body_count)
            self._rigid_body_graph[self._rigid_body_count].append(current_rigid_body_count)
            self.walk_rigid_body_graph(neigh)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#

import argparse
import gzip
import os
import sys
import warnings

from rdkit import Chem

from meeko import PDBQTMolecule
from meeko import RDKitMolCreate

try:
    from openbabel import openbabel as ob
    from meeko import obutils # fails if openbabel not available
except:
    _has_openbabel = False
else:
    _has_openbabel = True


def cmd_lineparser():
    parser = argparse.ArgumentParser(description='Export docking results to SDF format')
    parser.add_argument(dest='docking_results_filename', nargs = "+",
                        action='store', help='Docking output file, either a PDBQT \
                        file from Vina or a DLG file from AD-GPU.')
    parser.add_argument('-i', '--original_input', dest='template_filename',
                        action='store', help='[not recommended] Template molecule file, i.e. the original \
                        file that was used to prepare the PDBQT filename (hopefully SDF). \
                        This option requires OpenBabel, and is kept only for backwards compatibility with \
                        meeko v0.2.* and older, because it is only since v0.3.* that the SMILES string is \
                        added to a PDBQT remark and used to initialize an RDKit molecule.')
    parser.add_argument('-o', '--output_filename', dest='output_filename',
                        action='store', help='Output molecule filename. If not specified, suffix _docked is \
                        added to the filename based on the input molecule file, and using the same \
                        molecule file format')
    parser.add_argument('-s', '--suffix', dest='suffix_name', default='_docked',
                        action='store', help='Add suffix to output filename if -o/--output_filename \
                        not specified. WARNING: If specified as empty string (\'\'), this will overwrite \
                        the original molecule input file (default: _docked).')
    parser.add_argument('-c', '--only_cluster_leads', action='store_true',
                        help='Keep top pose from each AutoDock-GPU cluster, sorted by \
                        predicted free energy of cluster lead.')
    parser.add_argument('-', '--',  dest='redirect_stdout', action='store_true',
                        help='do not write file, redirect output to STDOUT. Arguments -o/--output_filename \
                        is ignored.')
    return parser.parse_args()

def main():
    args = cmd_lineparser()

    docking_results_filenames = args.docking_results_filename
    template_filename = args.template_filename
    output_filename = args.output_filename
    suffix_name = args.suffix_name
    only_cluster_leads = args.only_cluster_leads
    redirect_stdout = args.redirect_stdout

    if output_filename is not None and len(docking_results_filenames) > 1:
        print("Option -o/--output_filename incompatible with multiple input files", file=sys.stderr)
        sys.exit()
    
    for filename in docking_results_filenames:
        is_dlg = filename.endswith('.dlg') or filename.endswith(".dlg.gz")
        if filename.endswith(".gz"):
            with gzip.open(filename) as f:
                string = f.read().decode()
        else:
            with open(filename) as f:
                string = f.read()
        pdbqt_mol = PDBQTMolecule(string, is_dlg=is_dlg, skip_typing=True)

        if template_filename is not None: # OBMol from template_filename
            if not _has_openbabel:
                raise ImportError('-i/--original_input requires openbabel which is not available')
            if output_filename is not None:
                output_format = os.path.splitext(output_filename)[1][1:]
            else:
                output_format = os.path.splitext(template_filename)[1][1:]
            conv = ob.OBConversion()
            success = conv.SetOutFormat(output_format)
            if not success:
                raise RuntimeError('Input molecule file format %s not recognized by OpenBabel.' % output_format)
            ori_obmol = obutils.load_molecule_from_file(template_filename)
            for pose in pdbqt_mol:
                copy_obmol = ob.OBMol(ori_obmol) # connectivity may be corrupted by removing and adding Hs multiple times
                pose.copy_coordinates_to_obmol(copy_obmol)
                output_string = conv.WriteString(copy_obmol)
        else: # RDKit mol from SMILES in docking output PDBQT remarks
            output_string, failures = RDKitMolCreate.write_sd_string(
                    pdbqt_mol, only_cluster_leads=only_cluster_leads)
            output_format = 'sdf'
            for i in failures:
                warnings.warn("molecule %d not converted to RDKit/SD File" % i)
            if len(failures) == len(pdbqt_mol._atom_annotations["mol_index"]):
                msg = "\nCould not convert to RDKit. Maybe meeko was not used for preparing\n"
                msg += "the input PDBQT for docking, and the SMILES string is missing?\n"
                msg += "Except for standard protein sidechains, all ligands and flexible residues\n"
                msg += "require a REMARK SMILES line in the PDBQT, which is added automatically by meeko."
                raise RuntimeError(msg)
        if not redirect_stdout:
            if output_filename is None:
                fn = '%s%s.%s' % (os.path.splitext(filename)[0], suffix_name, output_format)
            else:
                fn = output_filename
            print(output_string, file=open(fn, 'w'))
        else:
            print(output_string)

    return

if __name__ == "__main__":
    sys.exit(main())


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Raccoon
# collection of useful snippets of code that's used frequently
#

import os
import importlib
from operator import itemgetter


mini_periodic_table = {
    1: "H",
    2: "He",
    3: "Li",
    4: "Be",
    5: "B",
    6: "C",
    7: "N",
    8: "O",
    9: "F",
    10: "Ne",
    11: "Na",
    12: "Mg",
    13: "Al",
    14: "Si",
    15: "P",
    16: "S",
    17: "Cl",
    18: "Ar",
    19: "K",
    20: "Ca",
    21: "Sc",
    22: "Ti",
    23: "V",
    24: "Cr",
    25: "Mn",
    26: "Fe",
    27: "Co",
    28: "Ni",
    29: "Cu",
    30: "Zn",
    31: "Ga",
    32: "Ge",
    33: "As",
    34: "Se",
    35: "Br",
    36: "Kr",
    37: "Rb",
    38: "Sr",
    39: "Y",
    40: "Zr",
    41: "Nb",
    42: "Mo",
    43: "Tc",
    44: "Ru",
    45: "Rh",
    46: "Pd",
    47: "Ag",
    48: "Cd",
    49: "In",
    50: "Sn",
    51: "Sb",
    52: "Te",
    53: "I",
    54: "Xe",
    55: "Cs",
    56: "Ba",
    57: "La",
    58: "Ce",
    59: "Pr",
    60: "Nd",
    61: "Pm",
    62: "Sm",
    63: "Eu",
    64: "Gd",
    65: "Tb",
    66: "Dy",
    67: "Ho",
    68: "Er",
    69: "Tm",
    70: "Yb",
    71: "Lu",
    72: "Hf",
    73: "Ta",
    74: "W",
    75: "Re",
    76: "Os",
    77: "Ir",
    78: "Pt",
    79: "Au",
    80: "Hg",
    81: "Tl",
    82: "Pb",
    83: "Bi",
    84: "Po",
    85: "At",
    86: "Rn",
    87: "Fr",
    88: "Ra",
}


def path_module(module_name):
    try:
        from importlib import util

        specs = util.find_spec(module_name)
        if specs is not None:
            return specs.submodule_search_locations[0]
    except:
        try:
            _, path, _ = imp.find_module(module_name)
            abspath = os.path.abspath(path)
            return abspath
        except ImportError:
            return None
    return None


def getNameExt(fname):
    """extract name and extension from the input file, removing the dot
    filename.ext -> [filename, ext]
    """
    name, ext = os.path.splitext(fname)
    return name, ext[1:]  # .lower()


class HJKRingDetection:
    """Implementation of the Hanser-Jauffret-Kaufmann exhaustive ring detection
    algorithm:
        ref:
        Th. Hanser, Ph. Jauffret, and G. Kaufmann
        J. Chem. Inf. Comput. Sci. 1996, 36, 1146-1152
    """

    def __init__(self, mgraph, max_iterations=8000000):
        self.mgraph = {key: [x for x in values] for (key, values) in mgraph.items()}
        self.rings = []
        self._iterations = 0
        self._max_iterations = max_iterations
        self._is_failed = False

    def scan(self, keep_chorded_rings=False, keep_equivalent_rings=False):
        """run the full protocol for exhaustive ring detection
        by default, only chordless rings are kept, and equivalent rings removed.
        (equivalent rings are rings that have the same size and share the same
        neighbors)
        """
        self.prune()
        self.build_pgraph()
        self.vertices = self._get_sorted_vertices()
        while self.vertices:
            self._remove_vertex(self.vertices[0])
        if not keep_chorded_rings:
            self.find_chordless_rings(keep_equivalent_rings)
        output_rings = []
        for ring in self.rings:
            output_rings.append(tuple(ring[:-1]))
        return output_rings

    def _get_sorted_vertices(self):
        """function to return the vertices to be removed, sorted by increasing
        connectivity order (see paper)"""
        vertices = ((k, len(v)) for k, v in self.mgraph.items())
        return [x[0] for x in sorted(vertices, key=itemgetter(1))]

    def prune(self):
        """iteratively prune graph until there are no leafs left (nodes with only
        one connection)"""
        while True:
            prune = []
            for node, neighbors in self.mgraph.items():
                if len(neighbors) == 1:
                    prune.append((node, neighbors))
            if len(prune) == 0:
                break
            for node, neighbors in prune:
                self.mgraph.pop(node)
                for n in neighbors:
                    self.mgraph[n].remove(node)

    def build_pgraph(self, prune=True):
        """convert the M-graph (molecular graph) into the P-graph (path/bond graph)"""
        self.pgraph = []
        for node, neigh in self.mgraph.items():
            for n in neigh:
                # use sets for unique id
                edge = set((node, n))
                if not edge in self.pgraph:
                    self.pgraph.append(edge)
        # re-convert the edges to lists because order matters in cycle detection
        self.pgraph = [list(x) for x in self.pgraph]

    def _remove_vertex(self, vertex):
        """remove a vertex and join all edges connected by that vertex (this is
        the REMOVE function from the paper)
        """
        visited = {}
        remove = []
        pool = []
        for path in self.pgraph:
            if self._has_vertex(vertex, path):
                pool.append(path)
        for i, path1 in enumerate(pool):
            for j, path2 in enumerate(pool):
                if i == j:
                    continue
                self._iterations += 1
                if self._iterations > self._max_iterations:
                    self._is_failed = True
                    break
                pair_id = tuple(set((i, j)))
                if pair_id in visited:
                    continue
                visited[pair_id] = None
                common = list(set(path1) & set(path2))
                common_count = len(common)
                # check if two paths have only this vertex in common or (or
                # two, if they're a cycle)
                if not 1 <= common_count <= 2:
                    continue
                # generate the joint path
                joint_path = self._concatenate_path(path1, path2, vertex)
                is_ring = joint_path[0] == joint_path[-1]
                # if paths share more than two vertices but they're not a ring, then skip
                if (common_count == 2) and not is_ring:
                    continue
                # store the ring...
                if is_ring:
                    self._add_ring(joint_path)
                # ...or the common path
                elif not joint_path in self.pgraph:
                    self.pgraph.append(joint_path)
        # remove used paths
        for p in pool:
            self.pgraph.remove(p)
        # remove the used vertex
        self.vertices.remove(vertex)

    def _add_ring(self, ring):
        """add newly found rings to the list (if not already there)"""
        r = set(ring)
        for candidate in self.rings:
            if r == set(candidate):
                return
        self.rings.append(ring)

    def _has_vertex(self, vertex, edge):
        """check if the vertex is part of this edge, and if true, return the
        sorted edge so that the vertex is the first in the list"""
        if edge[0] == vertex:
            return edge
        if edge[-1] == vertex:
            return edge[::-1]
        return None

    def _concatenate_path(self, path1, path2, v):
        """concatenate two paths sharing a common vertex
        a-b, c-b => a-b-c : idx1=1, idx2=1
        b-a, c-b => a-b-c : idx1=0, idx2=1
        a-b, b-c => a-b-c : idx1=1, idx2=0
        b-a, b-c => a-b-c : idx1=0, idx2=0
        """
        if not path1[-1] == v:
            path1.reverse()
        if not path2[0] == v:
            path2.reverse()
        return path1 + path2[1:]

    def _edge_in_pgraph(self, edge):
        """check if edge is already in pgraph"""
        e = set(edge)
        for p in self.pgraph:
            if e == set(p) and len(p) == len(edge):
                return True
        return False

    def find_chordless_rings(self, keep_equivalent_rings):
        """find chordless rings: cycles in which two vertices are not connected
        by an edge that does not itself belong to the cycle (Source:
        https://en.wikipedia.org/wiki/Cycle_%28graph_theory%29#Chordless_cycle)

        - iterate through rings starting from the smallest ones: A,B,C,D...
        - for each ring (A), find a candidate (e.g.: B) that is smaller and shares at least an edge
        - for this pair, calculate the two differences (A-B and B-A) in the list of edges of each
        - if  ( (A-B) + (B-A) ) a smaller ring (e.g.: C), then the current ring has a chord
        """
        # sort rings by the smallest to largest
        self.rings.sort(key=len, reverse=False)
        chordless_rings = []
        ring_edges = []
        rings_set = [set(x) for x in self.rings]
        for r in self.rings:
            edges = []
            for i in range(len(r) - 1):
                edges.append(
                    tuple(
                        set((r[i], r[(i + 1) % len(r)])),
                    )
                )
            edges = sorted(edges, key=itemgetter(0))
            ring_edges.append(edges)
        ring_contacts = {}
        for i, r1 in enumerate(self.rings):
            chordless = True
            r1_edges = ring_edges[i]
            ring_contacts[i] = []
            for j, r2 in enumerate(self.rings):
                if i == j:
                    continue
                if len(r2) >= len(r1):
                    # the candidate ring is larger than or the same size of the candidate
                    continue
                # avoid rings that don't share at least an edge
                # shared = set(r1) & set(r2)
                r2_edges = ring_edges[j]
                shared = set(r1_edges) & set(r2_edges)
                if len(shared) < 1:
                    continue
                ring_contacts[i].append(j)
                # get edges difference (r2_edges - r1_edges)
                core_edges = [x for x in r2_edges if not x in r1_edges]
                chord = [x for x in r1_edges if not x in r2_edges]
                # combined = chord + core_edges
                ring_new = []
                for edge in chord + core_edges:
                    ring_new.append(edge[0])
                    ring_new.append(edge[1])
                ring_new = set(ring_new)
                if (ring_new in rings_set) and (len(ring_new) < len(r1) - 1):
                    chordless = False
                    break
            if chordless:
                chordless_rings.append(i)
                ring_contacts[i] = set(ring_contacts[i])
        if not keep_equivalent_rings:
            chordless_rings = self._remove_equivalent_rings(
                chordless_rings, ring_contacts
            )
        self.rings = [self.rings[x] for x in chordless_rings]
        return

    def _remove_equivalent_rings(self, chordless_rings, ring_contacts):
        """remove equivalent rings by clustering by size, then by ring neighbors.
        Two rings A and B are equivalent if satisfy the following conditions:
            - same size
            - same neighbor ring(s) [C,D, ...]
            - (A - C) == (B -C)
        """
        size_clusters = {}
        # cluster rings by their size
        for ring_id in chordless_rings:
            if len(ring_contacts[ring_id]) == 0:
                continue
            size = len(self.rings[ring_id]) - 1
            if not size in size_clusters:
                size_clusters[size] = []
            size_clusters[size].append(ring_id)
        remove = []
        # process rings of the same size
        for size, ring_pool in size_clusters.items():
            for ri in ring_pool:
                if ri in remove:
                    continue
                for rj in ring_pool:
                    if ri == rj:
                        continue
                    common_neigh = ring_contacts[ri] & ring_contacts[rj]
                    for c in common_neigh:
                        d1 = set(self.rings[ri]) - set(self.rings[c])
                        d2 = set(self.rings[rj]) - set(self.rings[c])
                        if d1 == d2:
                            remove.append(rj)
        chordless_rings = [i for i in chordless_rings if not i in set(remove)]
        # for r in set(remove):
        #    chordless_rings.remove(r)
        return chordless_rings


# def writeList(filename, inlist, mode = 'w', addNewLine = False):
#     if addNewLine: nl = "\n"
#     else: nl = ""
#     fp = open(filename, mode)
#     for i in inlist:
#         fp.write(str(i)+nl)
#     fp.close()


# def getResInfo(string):
#    """ CHAIN:RESnum -> [ "CHAIN", "RES", num ]"""
#    if ':' in string:
#        chain, resraw = string.split(':')
#    else:
#        chain = ''
#        resraw = string
#    try:
#        res = resraw[0:3]
#        num = int(resraw[3:])
#    except:
#        # heuristic for nucleic acids
#        regex = r'[UACGT]+\d'
#        match = re.search(regex, resraw)
#        if match is None:
#            print("WARNING! Unknown residue naming scheme")
#            return chain, "X", "X"
#        res = resraw[0]
#        num = int(resraw[1:])
#        #print "NUCLEIC:",  chain, res, num
#    return chain, res, num


# def get_data_file(file_handle, dir_name, data_file):
#     module_dir, module_fname = os.path.split(file_handle)
#     DATAPATH = os.path.join(module_dir, dir_name, data_file)
#     return DATAPATH



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko
#

import os
from collections import defaultdict

import numpy as np
from scipy import spatial

from .utils.covalent_radius_table import covalent_radius
from .utils.autodock4_atom_types_elements import autodock4_atom_types_elements

try:
    from openbabel import openbabel as ob
except ImportError:
    _has_openbabel = False
else:
    _has_openbabel = True


atom_property_definitions = {'H': 'vdw', 'C': 'vdw', 'A': 'vdw', 'N': 'vdw', 'P': 'vdw', 'S': 'vdw',
                             'Br': 'vdw', 'I': 'vdw', 'F': 'vdw', 'Cl': 'vdw',
                             'NA': 'hb_acc', 'OA': 'hb_acc', 'SA': 'hb_acc', 'OS': 'hb_acc', 'NS': 'hb_acc',
                             'HD': 'hb_don', 'HS': 'hb_don',
                             'Mg': 'metal', 'Ca': 'metal', 'Fe': 'metal', 'Zn': 'metal', 'Mn': 'metal',
                             'MG': 'metal', 'CA': 'metal', 'FE': 'metal', 'ZN': 'metal', 'MN': 'metal',
                             'W': 'water',
                             'G0': 'glue', 'G1': 'glue', 'G2': 'glue', 'G3': 'glue',
                             'CG0': 'glue', 'CG1': 'glue', 'CG2': 'glue', 'CG3': 'glue'}


def _read_ligand_pdbqt_file(pdbqt_string, poses_to_read=-1, energy_range=-1, is_dlg=False, skip_typing=False):
    i = 0
    n_poses = 0
    previous_serial = 0
    tmp_positions = []
    tmp_atoms = []
    tmp_actives = []
    tmp_pdbqt_string = ''
    water_indices = {*()}  
    location = 'ligand'
    energy_best_pose = None
    is_first_pose = True
    is_model = False
    mol_index = -1 # incremented for each ROOT keyword
    atoms_dtype = [('idx', 'i4'), ('serial', 'i4'), ('name', 'U4'), ('resid', 'i4'),
                   ('resname', 'U3'), ('chain', 'U1'), ('xyz', 'f4', (3)),
                   ('partial_charges', 'f4'), ('atom_type', 'U3')]

    atoms = None
    positions = []

    # flexible_residue is for atoms between BEGIN_RES and END_RES keywords, ligand otherwise.
    # flexres assigned "ligand" if BEGIN/END RES keywords are missing
    # mol_index distinguishes different ligands and flexres because ROOT keyword increments mol_index  
    atom_annotations = {'ligand': [], 'flexible_residue': [], 'water': [],
                        'hb_acc': [], 'hb_don': [],
                        'all': [], 'vdw': [],
                        'glue': [], 'reactive': [], 'metal': [],
                        'mol_index': {},
                        }
    pose_data = {
        'n_poses': None,
        'active_atoms': [],
        'free_energies': [], 
        'intermolecular_energies': [],
        'internal_energies': [],
        'index_map': {},
        'pdbqt_string': [],
        'smiles': {},
        'smiles_index_map': {},
        'smiles_h_parent': {},
        'cluster_id': [],
        'rank_in_cluster': [],
        'cluster_leads_sorted': [],
        'cluster_size': [],
    }

    tmp_cluster_data = {}

    buffer_index_map = {}
    buffer_smiles = None
    buffer_smiles_index_map = []
    buffer_smiles_h_parent = []

    lines = pdbqt_string.split('\n')
    if len(lines[-1]) == 0: lines = lines[:-1]
    lines = [line + '\n' for line in lines]
    for line in lines:
        if is_dlg:
            if line.startswith('DOCKED'):
                line = line[8:]
            # parse clustering
            elif line.endswith('RANKING\n'):
                fields = line.split()
                cluster_id = int(fields[0])
                subrank = int(fields[1])
                run_id = int(fields[2])
                tmp_cluster_data[run_id] = (cluster_id, subrank)
            else:
                continue

        if not line.startswith(('MODEL', 'ENDMDL')):
            """This is very lazy I know...
            But would you rather spend time on rebuilding the whole torsion tree and stuff
            for writing PDBQT files or drinking margarita? Energy was already spend to build
            that, so let's re-use it!"""
            tmp_pdbqt_string += line

        if line.startswith('MODEL'):
            # Reinitialize variables
            i = 0
            previous_serial = 0
            tmp_positions = []
            tmp_atoms = []
            tmp_actives = []
            tmp_pdbqt_string = ''
            is_model = True
            mol_index = -1 # incremented for each ROOT keyword
        elif line.startswith('ATOM') or line.startswith("HETATM"):
            serial = int(line[6:11].strip())
            name = line[12:16].strip()
            resname = line[17:20].strip()
            chainid = line[21].strip()
            resid = int(line[22:26].strip())
            xyz = np.array([line[30:38].strip(), line[38:46].strip(), line[46:54].strip()], dtype=float)
            try:
                # PDBQT files from dry.py script are stripped from their partial charges. sigh...
                partial_charges = float(line[71:77].strip())
            except:
                partial_charges = 0.0
            atom_type = line[77:-1].strip()

            """ We are looking for gap in the serial atom numbers. Usually if they
            are not following it means that atoms are missing. This will happen with
            water molecules after using dry.py, only non-overlapping water molecules
            are kept. Also if the current serial becomes suddenly inferior than the
            previous and equal to 1, it means that we are now in another molecule/flexible 
            residue. So here we are adding dummy atoms
            """
            if (previous_serial + 1 != serial) and not (serial < previous_serial and serial == 1):
                diff = serial - previous_serial - 1
                for _ in range(diff):
                    xyz_nan = [999.999, 999.999, 999.999]
                    tmp_atoms.append((i, 9999, 'XXXX', 9999, 'XXX', 'X', xyz_nan, 999.999, 'XX'))
                    tmp_positions.append(xyz_nan)
                    i += 1

            # Once it is done, we can return to a normal life... and add existing atoms
            tmp_atoms.append((i, serial, name, resid, resname, chainid, xyz, partial_charges, atom_type))
            tmp_positions.append(xyz)
            tmp_actives.append(i)

            if is_first_pose:
                atom_annotations["mol_index"].setdefault(mol_index, [])
                atom_annotations["mol_index"][mol_index].append(i)
                # We store water idx separately from the rest since their number can be variable
                if atom_type != 'W':
                    atom_annotations[location].append(i)
                    atom_annotations['all'].append(i)
                    if not skip_typing:
                        atom_annotations[atom_property_definitions[atom_type]].append(i)

            if atom_type == 'W':
                water_indices.update([i])

            previous_serial = serial
            i += 1
        elif line.startswith("ROOT") and is_first_pose:
            mol_index += 1
            # buffers needed because REMARKS preceeds ROOT
            pose_data["index_map"][mol_index] = buffer_index_map
            pose_data["smiles"][mol_index] = buffer_smiles
            pose_data["smiles_index_map"][mol_index] = buffer_smiles_index_map
            pose_data["smiles_h_parent"][mol_index] = buffer_smiles_h_parent
            buffer_index_map = {}
            buffer_smiles = None
            buffer_smiles_index_map = []
            buffer_smiles_h_parent = []
        elif line.startswith('REMARK INDEX MAP') and is_first_pose:
            integers = [int(integer) for integer in line.split()[3:]]
            if len(integers) % 2 == 1:
                raise RuntimeError("Number of indices in INDEX MAP is odd")
            for j in range(int(len(integers) / 2)): 
                buffer_index_map[integers[j*2]] = integers[j*2 + 1]
        elif line.startswith('REMARK SMILES IDX') and is_first_pose:
            integers = [int(integer) for integer in line.split()[3:]]
            if len(integers) % 2 == 1:
                raise RuntimeError("Number of indices in SMILES IDX is odd")
            buffer_smiles_index_map.extend(integers)
        elif line.startswith('REMARK H PARENT') and is_first_pose:
            integers = [int(integer) for integer in line.split()[3:]]
            if len(integers) % 2 == 1:
                raise RuntimeError("Number of indices in H PARENT is odd")
            buffer_smiles_h_parent.extend(integers)
        elif line.startswith('REMARK SMILES') and is_first_pose: # must check after SMILES IDX
            buffer_smiles = line.split()[2]
        elif line.startswith('REMARK VINA RESULT') or line.startswith('USER    Estimated Free Energy of Binding    ='):
            # Read free energy from output PDBQT files
            try:
                # Vina
                energy = float(line.split()[3])
            except:
                # AD4
                energy = float(line[45:].split()[0]) # no guarantee of space between = and number

            if energy_best_pose is None:
                energy_best_pose = energy
            energy_current_pose = energy

            # NOTE this assumes poses are sorted by increasing energy
            diff_energy = energy_current_pose - energy_best_pose
            if (energy_range <= diff_energy and energy_range != -1):
                break

            pose_data['free_energies'].append(energy)
        elif not is_dlg and line.startswith('REMARK INTER:'):
            pose_data['intermolecular_energies'].append(float(line.split()[2]))
        elif not is_dlg and line.startswith('REMARK INTRA:'):
            pose_data['internal_energies'].append(float(line.split()[2]))
        elif is_dlg and line.startswith('USER    (1) Final Intermolecular Energy     ='):
            pose_data['intermolecular_energies'].append(float(line[45:].split()[0]))
        elif is_dlg and line.startswith('USER    (2) Final Total Internal Energy     ='):
            pose_data['internal_energies'].append(float(line[45:].split()[0]))
        elif line.startswith('BEGIN_RES'):
            location = 'flexible_residue'
        elif line.startswith('END_RES'):
            # We never know if there is a molecule just after the flexible residue...
            location = 'ligand'
        elif line.startswith('ENDMDL'):
            n_poses += 1
            # After reading the first pose no need to store atom properties
            # anymore, it is the same for every pose
            is_first_pose = False

            tmp_atoms = np.array(tmp_atoms, dtype=atoms_dtype)

            if atoms is None:
                """We store the atoms (topology) only once, since it is supposed to be
                the same for all the molecules in the PDBQT file (except when water molecules
                are involved... classic). But we will continue to compare the topology of
                the current pose with the first one seen in the PDBQT file, to be sure only
                the atom positions are changing."""
                atoms = tmp_atoms.copy()
            else:
                # Check if the molecule topology is the same for each pose
                # We ignore water molecules (W) and atom type XX
                columns = ['idx', 'serial', 'name', 'resid', 'resname', 'chain', 'partial_charges', 'atom_type']
                topology1 = atoms[np.isin(atoms['atom_type'], ['W', 'XX'], invert=True)][columns]
                topology2 = tmp_atoms[np.isin(atoms['atom_type'], ['W', 'XX'], invert=True)][columns]

                if not np.array_equal(topology1, topology2):
                    error_msg = 'molecules have different topologies'
                    raise RuntimeError(error_msg)

                # Update information about water molecules (W) as soon as we find new ones
                tmp_water_molecules_idx = tmp_atoms[tmp_atoms['atom_type'] == 'W']['idx']
                water_molecules_idx = atoms[atoms['atom_type'] == 'XX']['idx']
                new_water_molecules_idx = list(set(tmp_water_molecules_idx).intersection(water_molecules_idx))
                atoms[new_water_molecules_idx] = tmp_atoms[new_water_molecules_idx]

            positions.append(tmp_positions)
            pose_data['active_atoms'].append(tmp_actives)
            pose_data['pdbqt_string'].append(tmp_pdbqt_string)

            if (n_poses >= poses_to_read and poses_to_read != -1):
                break

    """ if there is no model, it means that there is only one molecule
    so when we reach the end of the file, we store the atoms, 
    positions and actives stuff. """
    if not is_model:
        n_poses += 1
        atoms = np.array(tmp_atoms, dtype=atoms_dtype)
        positions.append(tmp_positions)
        pose_data['active_atoms'].append(tmp_actives)
        pose_data['pdbqt_string'].append(tmp_pdbqt_string)

    positions = np.array(positions).reshape((n_poses, atoms.shape[0], 3))

    pose_data['n_poses'] = n_poses

    # We add indices of all the water molecules we saw
    if water_indices:
        atom_annotations['water'] = list(water_indices)

    # clustering        
    if len(tmp_cluster_data) > 0:
        if len(tmp_cluster_data) != n_poses:
            raise RuntimeError("Nr of poses in cluster data (%d) differs from nr of poses (%d)" % (len(tmp_cluster_data, n_poses)))
        pose_data["cluster_id"] = [None] * n_poses
        pose_data["rank_in_cluster"] = [None] * n_poses
        pose_data["cluster_size"] = [None] * n_poses
        cluster_ids = [cluster_id for _, (cluster_id, _) in tmp_cluster_data.items()]
        n_clusters = max(cluster_ids)
        pose_data["cluster_leads_sorted"] = [None] * n_clusters
        for pose_index, (cluster_id, rank_in_cluster) in tmp_cluster_data.items():
            pose_data["cluster_id"][pose_index - 1] = cluster_id
            pose_data["rank_in_cluster"][pose_index - 1] = rank_in_cluster
            pose_data["cluster_size"][pose_index - 1] = cluster_ids.count(cluster_id)
            if rank_in_cluster == 1: # is cluster lead
                pose_data["cluster_leads_sorted"][cluster_id - 1] = pose_index - 1
    return atoms, positions, atom_annotations, pose_data


def _identify_bonds(atom_idx, positions, atom_types):
    bonds = defaultdict(list)
    KDTree = spatial.cKDTree(positions)
    bond_allowance_factor = 1.1
    # If we ask more than the number of coordinates/element
    # in the BHTree, we will end up with some inf values
    k = 5 if len(atom_idx) > 5 else len(atom_idx)
    atom_idx = np.array(atom_idx)

    # If there is only one atom, we know there won't be a single bond..
    if len(atom_idx) == 1:
        return bonds

    for atom_i, position, atom_type in zip(atom_idx, positions, atom_types):
        distances, indices = KDTree.query(position, k=k)
        r_cov = covalent_radius[autodock4_atom_types_elements[atom_type]]

        optimal_distances = [bond_allowance_factor * (r_cov + covalent_radius[autodock4_atom_types_elements[atom_types[i]]]) for i in indices[1:]]
        bonds[atom_i] = atom_idx[indices[1:][np.where(distances[1:] < optimal_distances)]].tolist()

    return bonds


class PDBQTMolecule:

    def __init__(self, pdbqt_string, name=None, poses_to_read=None, energy_range=None, is_dlg=False, skip_typing=False):
        """PDBQTMolecule class for reading PDBQT (or dlg) files from AutoDock4, AutoDock-GPU or AutoDock-Vina

        Contains both __getitem__ and __iter__ methods, someone might lose his mind because of this.

        Args:
            pdbqt_string (str): pdbqt string
            name (str): name of the molecule (default: None, use filename without pdbqt suffix)
            poses_to_read (int): total number of poses to read (default: None, read all)
            energy_range (float): read docked poses until the maximum energy difference 
                from best pose is reach, for example 2.5 kcal/mol (default: Non, read all)
            is_dlg (bool): input file is in dlg (AutoDock docking log) format (default: False)
            skip_typing (bool, optional): Flag indicating that atomtyping should be skipped
        """
        self._current_pose = 0
        self._pdbqt_filename = None
        self._atoms = None
        self._positions = None
        self._bonds = None
        self._atom_annotations = None
        self._pose_data = None
        self._name = name

        # Juice all the information from that PDBQT file
        poses_to_read = poses_to_read if poses_to_read is not None else -1
        energy_range = energy_range if energy_range is not None else -1
        results = _read_ligand_pdbqt_file(pdbqt_string, poses_to_read, energy_range, is_dlg, skip_typing)
        self._atoms, self._positions, self._atom_annotations, self._pose_data = results

        if self._atoms.shape[0] == 0:
            raise RuntimeError('read 0 atoms. Consider PDBQTMolecule.from_file(fname)')

        # Build KDTrees for each pose (search closest atoms by distance)
        self._KDTrees = [spatial.cKDTree(positions) for positions in self._positions]

        # Identify bonds in the ligands
        if not skip_typing:
            mol_atoms = self._atoms[self._atom_annotations['ligand']]
            self._bonds = _identify_bonds(self._atom_annotations['ligand'], mol_atoms['xyz'], mol_atoms['atom_type'])

            """... then in the flexible residues 
            Since we are extracting bonds from docked poses, we might be in the situation
            where the ligand reacted with one the flexible residues and we don't want to 
            consider them as normally bonded..."""
            if self.has_flexible_residues():
                flex_atoms = self._atoms[self._atom_annotations['flexible_residue']]
                self._bonds.update(_identify_bonds(self._atom_annotations['flexible_residue'], flex_atoms['xyz'], flex_atoms['atom_type']))

    @classmethod
    def from_file(cls, pdbqt_filename, name=None, poses_to_read=None, energy_range=None, is_dlg=False, skip_typing=False): 
        if name is None:
            name = os.path.splitext(os.path.basename(pdbqt_filename))[0]
        with open(pdbqt_filename) as f:
            pdbqt_string = f.read()
        instance = cls(pdbqt_string, name, poses_to_read, energy_range, is_dlg, skip_typing)
        instance._pdbqt_filename = pdbqt_filename
        return instance

    def __getitem__(self, value):
        if isinstance(value, int):
            if value < 0 or value >= self._positions.shape[0]:
                raise IndexError('The index (%d) is out of range.' % value)
        elif isinstance(value, slice):
            raise TypeError('Slicing is not implemented for PDBQTMolecule object.')
        else:
            raise TypeError('Invalid argument type.')

        self._current_pose = value
        return self

    def __iter__(self):
        self._current_pose = -1
        return self

    def __next__(self):
        if self._current_pose + 1 >= self._positions.shape[0]:
            raise StopIteration

        self._current_pose += 1

        return self

    def __repr__(self):
        repr_str = '<Molecule named %s containing %d poses of %d atoms>'
        return (repr_str % (self._name, self._pose_data['n_poses'], self._atoms.shape[0]))

    @property    
    def name(self):
        """Return the name of the molecule."""
        return self._name

    @property
    def pose_id(self):
        """Return the index of the current pose."""
        return self._current_pose

    @property
    def score(self):
        """Return the score (kcal/mol) of the current pose."""
        return self._pose_data['free_energies'][self._current_pose]

    def available_atom_properties(self, ignore_properties=None):
        """Return all the available atom properties for that molecule.

        The following properties are ignored: ligand and flexible_residue

        """
        if ignore_properties is None:
            ignore_properties = []

        if not isinstance(ignore_properties, (list, tuple)):
            ignore_properties = [ignore_properties]

        ignore_properties += ['ligand', 'flexible_residue', 'water']

        return [k for k, v in self._atom_annotations.items()
                if k not in ignore_properties and len(v) > 0]

    def has_flexible_residues(self):
        """Tell if the molecule contains a flexible residue or not.

        Returns:
            bool: True if contains flexible residues, otherwise False

        """
        if self._atom_annotations['flexible_residue']:
            return True
        else:
            return False

    def has_water_molecules(self):
        """Tell if the molecules contains water molecules or not in the current pose.

        Returns:
            bool: True if contains water molecules in the current pose, otherwise False

        """
        active_atoms_idx = self._pose_data['active_atoms'][self._current_pose]
        if set(self._atom_annotations['water']).intersection(active_atoms_idx):
            return True
        else:
            return False

    def atoms(self, atom_idx=None, only_active=True):
        """Return the atom i

        Args:
            atom_idx (int, list): index of one or multiple atoms (0-based)
            only_active (bool): return only active atoms (default: True, return only active atoms)

        Returns:
            ndarray: 2d ndarray (atom_id, atom_name, resname, resid, chainid, xyz, q, t)

        """
        if atom_idx is not None:
            if not isinstance(atom_idx, (list, tuple, np.ndarray)):
                atom_idx = np.array(atom_idx, dtype=np.int)
        else:
            atom_idx = np.arange(0, self._atoms.shape[0])

        # Get index of only the active atoms
        if only_active:
            active_atoms_idx = self._pose_data['active_atoms'][self._current_pose]
            atom_idx = sorted(list(set(atom_idx).intersection(active_atoms_idx)))

        atoms = self._atoms[atom_idx].copy()
        atoms['xyz'] = self._positions[self._current_pose, atom_idx,:]

        return atoms

    def positions(self, atom_idx=None, only_active=True):
        """Return coordinates (xyz) of all atoms or a certain atom

        Args:
            atom_idx (int, list): index of one or multiple atoms (0-based)
            only_active (bool): return only active atoms (default: True, return only active atoms)

        Returns:
            ndarray: 2d ndarray of coordinates (xyz)

        """
        return np.atleast_2d(self.atoms(atom_idx, only_active)['xyz'])

    def atoms_by_properties(self, atom_properties, only_active=True):
        """Return atom based on their properties

        Args:
            atom_properties (str or list): property of the atoms to retrieve 
                (properties: ligand, flexible_residue, vdw, hb_don, hb_acc, metal, water, reactive, glue)
            only_active (bool): return only active atoms (default: True, return only active atoms)

        """
        if not isinstance(atom_properties, (list, tuple)):
            atom_properties = [atom_properties]

        if len(atom_properties) > 1:
            try:
                atom_idx = set(self._atom_annotations[atom_properties[0]])

                for atom_property in atom_properties[1:]:
                    atom_idx.intersection_update(self._atom_annotations[atom_property])
            except:
                error_msg = 'Atom property %s is not valid. Valid atom properties are: %s'
                raise KeyError(error_msg % (atom_property, self._atom_annotations.keys()))

            atom_idx = list(atom_idx)
        else:
            try:
                atom_idx = self._atom_annotations[atom_properties[0]]
            except:
                error_msg = 'Atom property %s is not valid. Valid atom properties are: %s'
                raise KeyError(error_msg % (atom_properties[0], self._atom_annotations.keys()))

        if atom_idx:
            return self.atoms(atom_idx, only_active)
        else:
            return np.array([])

    def closest_atoms_from_positions(self, xyz, radius, atom_properties=None, ignore=None):
        """Retrieve indices of the closest atoms around a positions/coordinates 
        at a certain radius.

        Args:
            xyz (np.ndarray): array of 3D coordinates
            raidus (float): radius
            atom_properties (str): property of the atoms to retrieve 
                (properties: ligand, flexible_residue, vdw, hb_don, hb_acc, metal, water, reactive, glue)
            ignore (int or list): ignore atom for the search using atom id (0-based)

        Returns:
            ndarray: 2d ndarray (atom_id, atom_name, resname, resid, chainid, xyz, q, t)

        """
        atom_idx = self._KDTrees[self._current_pose].query_ball_point(xyz, radius, p=2, return_sorted=True)

        # When nothing was found around...
        if not atom_idx:
            return np.array([])

        # Handle the case when positions for of only one atom was passed in the input
        try:
            atom_idx = {i for j in atom_idx for i in j}
        except:
            atom_idx = set(atom_idx)

        if atom_properties is not None:
            if not isinstance(atom_properties, (list, tuple)):
                atom_properties = [atom_properties]

            try:
                for atom_property in atom_properties:
                    atom_idx.intersection_update(self._atom_annotations[atom_property])
            except:
                error_msg = 'Atom property %s is not valid. Valid atom properties are: %s'
                raise KeyError(error_msg % (atom_property, self._atom_annotations.keys()))

        if ignore is not None:
            if not isinstance(ignore, (list, tuple, np.ndarray)):
                ignore = [ignore]
            atom_idx = atom_idx.difference([i for i in ignore])

        # Get index of only the active atoms
        active_atoms_idx = self._pose_data['active_atoms'][self._current_pose]
        atom_idx = list(set(atom_idx).intersection(active_atoms_idx))

        if atom_idx:
            atoms = self._atoms[atom_idx].copy()
            atoms['xyz'] = self._positions[self._current_pose, atom_idx,:]
            return atoms
        else:
            return np.array([])

    def closest_atoms(self, atom_idx, radius, atom_properties=None):
        """Retrieve indices of the closest atoms around a positions/coordinates 
        at a certain radius.

        Args:
            atom_idx (int, list): index of one or multiple atoms (0-based)
            raidus (float): radius
            atom_properties (str or list): property of the atoms to retrieve 
                (properties: ligand, flexible_residue, vdw, hb_don, hb_acc, metal, water, reactive, glue)

        Returns:
            ndarray: ndarray (atom_id, atom_name, resname, resid, chainid, xyz, q, t)

        """
        if not isinstance(atom_idx, (list, tuple)):
                atom_idx = [atom_idx]

        # Get index of only the active atoms
        active_atoms_idx = self._pose_data['active_atoms'][self._current_pose]
        atom_idx = list(set(atom_idx).intersection(active_atoms_idx))

        if atom_idx:
            positions = self._positions[self._current_pose, atom_idx,:]
            return self.closest_atoms_from_positions(positions, radius, atom_properties, atom_idx)
        else:
            return np.array([])

    def neighbor_atoms(self, atom_idx):
        """Return neighbor (bonded) atoms

        Args:
            atom_idx (int, list): index of one or multiple atoms (0-based)

        Returns:
            list_of_list: list of lists containing the neighbor (bonded) atoms (0-based)

        """
        if not isinstance(atom_idx, (list, tuple, np.ndarray)):
            atom_idx = [atom_idx]

        # Get index of only the active atoms
        active_atoms_idx = self._pose_data['active_atoms'][self._current_pose]
        atom_idx = list(set(atom_idx).intersection(active_atoms_idx))

        return [self._bonds[i] for i in atom_idx]

    def write_pdbqt_string(self, as_model=True):
        """Write PDBQT output string of the current pose
        
        Args:
            as_model (bool): Qdd MODEL/ENDMDL keywords to the output PDBQT string (default: True)
        
        Returns:
            string: Description
        
        """
        if as_model:
            pdbqt_string = 'MODEL    %5d\n' % (self._current_pose + 1)
            pdbqt_string += self._pose_data['pdbqt_string'][self._current_pose]
            pdbqt_string += 'ENDMDL\n'
            return pdbqt_string
        else: 
            return self._pose_data['pdbqt_string'][self._current_pose]

    def write_pdbqt_file(self, output_pdbqtfilename, overwrite=False, as_model=False):
        """Write PDBQT file of the current pose

        Args:
            output_pdbqtfilename (str): filename of the output PDBQT file
            overwrite (bool): overwrite on existing pdbqt file (default: False)
            as_model (bool): Qdd MODEL/ENDMDL keywords to the output PDBQT string (default: False)

        """
        print(overwrite and os.path.isfile(output_pdbqtfilename))
        if not overwrite and os.path.isfile(output_pdbqtfilename):
            raise RuntimeError('Output PDBQT file %s already exists' % output_pdbqtfilename)

        if as_model:
            pdbqt_string = 'MODEL    %5d\n' % (self._current_pose + 1)
            pdbqt_string += self._pose_data['pdbqt_string'][self._current_pose] 
            pdbqt_string += 'ENDMDL\n'
        else:
            pdbqt_string = self._pose_data['pdbqt_string'][self._current_pose]

        with open(output_pdbqtfilename, 'w') as w:
            w.write(pdbqt_string)

    def copy_coordinates_to_obmol(self, obmol, index_map=None):
        """Copy coordinates of the current pose to an obmol object 

        Args:
            obmol (OBMol): coordinates will be changed in this object
            index_map (dict): map of atom indices from obmol (keys) to coords (values) (Default: None)

        """
        if not _has_openbabel:
            raise ImportError('openbabel is required')

        if index_map is None:
            index_map = self._pose_data['index_map']

        n_atoms = obmol.NumAtoms()
        n_matched_atoms = 0
        hydrogens_to_delete = []
        heavy_parents = []

        for atom in ob.OBMolAtomIter(obmol):
            ob_index = atom.GetIdx() # 1-index

            if ob_index in index_map:
                pdbqt_index = index_map[ob_index] - 1
                x, y, z = self._positions[self._current_pose][pdbqt_index, :]
                atom.SetVector(x, y, z)
                n_matched_atoms += 1
            elif atom.GetAtomicNum() != 1:
                raise RuntimeError('Heavy atom in OBMol is missing, only hydrogens can be missing')
            else:
                hydrogens_to_delete.append(atom)
                bond_counter = 0

                for bond in ob.OBAtomBondIter(atom):
                    bond_counter += 1
                if bond_counter != 1:
                    raise RuntimeError('Hydrogen atom has more than one bonds (%d bonds)' % bond_counter)

                begin_atom = bond.GetBeginAtom()
                end_atom = bond.GetEndAtom()

                if atom == begin_atom:
                    heavy_parents.append(end_atom)
                elif atom == end_atom:
                    heavy_parents.append(begin_atom)
                else:
                    raise RuntimeError('Hydrogen isn\'t either Begin or End atom of its own bond')

        if n_matched_atoms != len(index_map):
            raise RuntimeError('Not all the atoms were considered')

        # delete explicit hydrogens
        for hydrogen in hydrogens_to_delete:
            obmol.DeleteHydrogen(hydrogen)

        # increment implicit H count of heavy atom parents
        for heavy_parent in heavy_parents:
            n_implicit = heavy_parent.GetImplicitHCount()
            heavy_parent.SetImplicitHCount(n_implicit + 1)

        # add back explicit hydrogens
        obmol.AddHydrogens()
        if obmol.NumAtoms() != n_atoms:
            raise RuntimeError('Number of atoms changed after deleting and adding hydrogens')



#!/usr/bin/env python

import argparse
import json
import math
from os import linesep as os_linesep
import pathlib
import sys

from meeko import PDBQTReceptor
from meeko import reactive_typer
from meeko import get_reactive_config
from meeko import gridbox

path_to_this_script = pathlib.Path(__file__).resolve()

def parse_residue_string(string):
    """
    Args:
        string (str): Residue identifier in the format 'chain:resname:resnum'.

    Returns:
        tuple: Residue identifier components: (chain, resname, resnum).
        bool: Whether the parsing was successful or not.
        str: An error message describing any encountered parsing errors.

    Example:
        >>> parse_residue_string("A:ALA:42")
        (('A', 'ALA', 42), True, '')
    """

    ok = True
    err = ""
    res_id = (None, None, None)
    if string.count(":") != 2:
        ok = False
        err += "Need exacly two ':' but found %d in '%s'" % (string.count(":"), string) + os_linesep
        err += "Example: 'A:HIE:42'" + os_linesep
        return res_id, ok, err
    chain, resname, resnum = string.split(":")
    if len(chain) not in (0, 1):
        ok = False
        err += "chain must be 0 or 1 char but it is '%s' (%d chars) in '%s'" % (chain, len(chain), string) + os_linesep
    if len(resname) > 3:
        ok = False
        err += "resname must be max 3 characters long, but is '%s' in '%s'" % (resname, string) + os_linesep
    try:
        resnum = int(resnum)
    except:
        ok = False
        err += "resnum could not be converted to integer, it was '%s' in '%s'" % (resnum, string) + os_linesep
    if ok:
        res_id = (chain, resname, resnum)
    return res_id, ok, err

def parse_residue_string_and_name(string):
    """
    Example:
        >>> parse_residue_string_and_name("A:HIE:42:CE1")
        ({"res_id": ('A', 'ALA', 42), "name": "CE1"}, True, '')
    """
    ok = True
    err = ""
    output = {"res_id": (None, None, None),
              "name": None}
    if string.count(":") != 3:
        ok = False
        err += "Need three ':' but found %d in %s" % (string.count(":"), string) + os_linesep
        err += "Example: 'A:HIE:42:CE1'" + os_linesep
        return output, ok, err

    resid_string, name = string.rsplit(":", 1)
    output["name"] = name
    res_id, ok_, err_ = parse_residue_string(resid_string)
    output["res_id"] = res_id
    ok &= ok_
    err += err_
    return output, ok, err

def parse_resname_and_name(string):
    """
    Example:
        >>> parse_resname_and_name("HIE:CE1")
        (("HIE", "CE1"), True, "")
    """
    ok = True
    err = ""
    output = (None, None)
    if string.count(":") != 1:
        ok = False
        err += "Expected one ':' but found %d in %s" % (string.count(":"), string) + os_linesep
        err += "Example: 'HIE:CE1'" + os_linesep
        return output, ok, err
    resname, name = string.split(":")
    return (resname, name), ok, err




class TalkativeParser(argparse.ArgumentParser):
    def error(self, message):
        """overload to print_help for every error"""
        self.print_help()
        this_script = path_to_this_script.name
        print('\n%s: error: %s' % (this_script, message), file=sys.stderr)
        sys.exit(2)

def check(success, error_msg):
    if not success:
        print("Error: " + error_msg, file=sys.stderr)
        sys.exit(2)

def get_args():
    parser = TalkativeParser()
    parser.add_argument('--pdb', help="input can be PDBQT but charges and types will be reassigned")
    parser.add_argument('--pdbqt', help="keeps existing charges and types")
    parser.add_argument('-o', '--output_filename', required=True, help="adds _rigid/_flex with flexible residues. Always suffixes .pdbqt.")
    parser.add_argument('-f', '--flexres', action="append", default=[],
                        help="repeat flag for each residue, e.g: -f \" :LYS:42\" -f \"B:TYR:23\" and keep space for empty chain")
    parser.add_argument('-r', '--reactive_flexres', action="append", default=[],
                        help="same as --flexres but for reactive residues (max 8)")
    parser.add_argument('-g', '--reactive_name', action="append", default=[],
                        help="set name of reactive atom of a residue type, e.g: -g 'TRP:NE1'. Overridden by --reactive_name_specific")
    parser.add_argument('-s', '--reactive_name_specific', action="append", default=[],
                        help="set name of reactive atom for an individual residue, e.g: -s 'A:HIE:42:NE2'. Residue will be reactive.")
    #parser.add_argument('-b', '--gridbox_filename', help="set grid box size and center using a Vina configuration file")
    parser.add_argument('--box_size', help="size of grid box (x, y, z) in Angstrom", nargs=3, type=float)
    parser.add_argument('--box_center', help="center of grid box (x, y, z) in Angstrom", nargs=3, type=float)
    parser.add_argument('--box_center_on_reactive_res', help="project center of grid box along CA-CB bond 5 A away from CB", action="store_true")
    parser.add_argument('--ligand', help="PDBQT of reference ligand")
    parser.add_argument('--padding', help="padding around reference ligand [A]", type=float)
    parser.add_argument('--skip_gpf', help="do not write a GPF file for autogrid", action="store_true")
    parser.add_argument('--r_eq_12', default=1.8, type=float, help="r_eq for reactive atoms (1-2 interaction)")
    parser.add_argument('--eps_12', default=2.5, type=float, help="epsilon for reactive atoms (1-2 interaction)")
    parser.add_argument('--r_eq_13_scaling', default=0.5, type=float, help="r_eq scaling for 1-3 interaction across reactive atoms")
    parser.add_argument('--r_eq_14_scaling', default=0.5, type=float, help="r_eq scaling for 1-4 interaction across reactive atoms")
    args = parser.parse_args()

    if (args.pdb is None) == (args.pdbqt is None):
        msg = "need either --pdb or --pdbqt"
        print("Command line error: " + msg, file=sys.stderr)
        sys.exit(2)
    if (args.box_center is not None) and args.box_center_on_reactive_res:
        msg = "can't use both --box_center and --box_center_on_reactive_res"
        print("Command line error: " + msg, file=sys.stderr)
        sys.exit(2)
    got_center = (args.box_center is not None) or args.box_center_on_reactive_res or (args.ligand is not None)
    if not args.skip_gpf:
        if not got_center:
            msg  = "missing center or size of grid box to write .gpf file for autogrid4" + os_linesep
            msg += "use --box_size and either --box_center or --box_center_on_reactive_res" + os_linesep
            msg += "or --ligand and --padding" + os_linesep
            msg += "Exactly one reactive residue required for --box_center_on_reactive_res" + os_linesep
            msg += "If a GPF file is not needed (e.g. docking with Vina scoring function) use option --skip_gpf"
            print("Command line error: " + msg, file=sys.stderr)
            sys.exit(2)
        if (args.box_size is None) and (args.padding is None):
            msg  = "grid box information is needed to dock with the AD4 scoring function." + os_linesep
            msg += "The grid box center and size will be used to write a GPF file for autogrid" + os_linesep
            msg += "If a GPF file is not needed (e.g. docking with Vina scoring function) use option --skip_gpf"
            print("Command line error: " + msg, file=sys.stderr)
            sys.exit(2)

    if (args.box_center is not None) + (args.ligand is not None) + args.box_center_on_reactive_res > 1:
        msg = "--box_center, --box_center_on_reactive_res, and --ligand are mutually exclusive options"
        print("Command line error: " + msg, file=sys.stderr)
        sys.exit(2)

    return args

def main():
    args = get_args()

    reactive_atom = {
        "SER":  "OG", "LYS":  "NZ", "TYR":  "OH", "CYS":  "SG", "HIE": "NE2",
        "HID": "ND1", "GLU": "OE2", "THR": "OG1", "MET":  "SD",
    }
    modified_resnames = set()
    for react_name_str in args.reactive_name:
        (resname, name), ok, err = parse_resname_and_name(react_name_str)
        if ok:
            if resname in modified_resnames:
                print("Command line error: repeated resname %s passed to --reactive_resname" % resname + os_linesep, file=sys.stderr)
                sys.exit(2)
            modified_resnames.add(resname)
            reactive_atom[resname] = name
        else:
            print("Error in parsing --reactive_name argument" + os_linesep, file=sys.stderr)
            print(err, file=sys.stderr)
            sys.exit(2)

    reactive_flexres = {}
    all_ok = True
    all_err = ""
    for resid_string in args.reactive_flexres:
        res_id, ok, err = parse_residue_string(resid_string)
        if ok:
            resname = res_id[1]
            if resname in reactive_atom:
                reactive_flexres[res_id] = reactive_atom[resname]
            else:
                all_ok = False
                all_err += "no default reactive name for %s, " % resname
                all_err += "use --reactive_name or --reactive_name_specific" + os_linesep
        all_ok &= ok
        all_err += err

    for string in args.reactive_name_specific:
        out, ok, err = parse_residue_string_and_name(string)
        if ok:
            # override name if res_id was also passed to --reactive_flexres
            reactive_flexres[out["res_id"]] = out["name"]
        all_ok &= ok
        all_err += err

    if len(reactive_flexres) > 8:
        msg = "got %d reactive_flexres but maximum is 8." % (len(args.reactive_flexres))
        print("Command line error: " + msg, file=sys.stderr)
        sys.exit(2)

    all_flexres = set()
    for resid_string in args.flexres:
        res_id, ok, err = parse_residue_string(resid_string)
        all_ok &= ok
        all_err += err
        if ok:
            if res_id in reactive_flexres:
                all_err += "Command line error: can't determine if %s is reactive or not." % str(res_id) + os_linesep
                all_err += "Do not pass %s to --flexres if it is reactive." % str(res_id) + os_linesep
                all_ok = False
            all_flexres.add(res_id)

    if all_ok:
        all_flexres = all_flexres.union(reactive_flexres)
    else:
        print("Error:", file=sys.stderr)
        print(all_err, file=sys.stderr)
        sys.exit(2)

    if len(all_flexres) > 0:
        print()
        print("Flexible residues:")
        print("chain resname resnum is_reactive reactive_atom")
        string = "%5s%8s%7d%12s%14s"
        for res_id in all_flexres:
            chain, resname, resnum = res_id
            is_react = res_id in reactive_flexres
            if is_react:
                react_atom = reactive_flexres[res_id]
            else:
                react_atom = ""
            print(string % (chain, resname, resnum, is_react, react_atom))
        print()

    if len(reactive_flexres) != 1 and args.box_center_on_reactive_res:
        msg = "--box_center_on-reactive_res can be used only with one reactive" + os_linesep
        msg += "residue, but %d reactive residues are set" % len(reactive_flexres)
        print("Command line error:" + msg, file=sys.stderr)
        sys.exit(2)

    if args.pdb is not None:
        receptor = PDBQTReceptor(args.pdb, skip_typing=True)
        ok, err = receptor.assign_types_charges()
        check(ok, err)
    else:
        receptor = PDBQTReceptor(args.pdbqt)

    any_lig_base_types = ["HD", "C", "A", "N", "NA", "OA", "F", "P", "SA",
                        "S", "Cl", "CL", "Br", "BR", "I", "Si", "B"]

    pdbqt, ok, err = receptor.write_pdbqt_string(flexres=all_flexres)
    check(ok, err)

    outpath = pathlib.Path(args.output_filename)

    written_files_log = {"filename": [], "description": []}

    if len(all_flexres) == 0:
        box_center = args.box_center
        rigid_fn = str(outpath.with_suffix(".pdbqt"))
        flex_fn = None
    else:
        all_flex_pdbqt = ""
        reactive_flexres_count = 0
        for res_id, flexres_pdbqt in pdbqt["flex"].items():
            if res_id in reactive_flexres:
                reactive_flexres_count += 1
                prefix_atype = "%d" % reactive_flexres_count
                resname = res_id[1]
                reactive_atom = reactive_flexres[res_id]
                flexres_pdbqt = PDBQTReceptor.make_flexres_reactive(flexres_pdbqt, reactive_atom, resname, prefix_atype)
            all_flex_pdbqt += flexres_pdbqt

        suffix = outpath.suffix
        if outpath.suffix == "":
            suffix = ".pdbqt"
        rigid_fn = str(outpath.with_suffix("")) + "_rigid" + suffix
        flex_fn = str(outpath.with_suffix("")) + "_flex" + suffix

        written_files_log["filename"].append(flex_fn)
        written_files_log["description"].append("flexible receptor input file")
        with open(flex_fn, "w") as f:
            f.write(all_flex_pdbqt)

    written_files_log["filename"].append(rigid_fn)
    written_files_log["description"].append("static (i.e., rigid) receptor input file")
    with open(rigid_fn, "w") as f:
        f.write(pdbqt["rigid"])

    # GPF for autogrid4
    if not args.skip_gpf:
        if args.box_center is not None:
            box_center = args.box_center
            box_size = args.box_size
        elif args.box_center_on_reactive_res:
            # we have only one reactive residue and will set the box center
            # to be 5 Angstromg away from CB along the CA->CB vector
            idxs = receptor.atom_idxs_by_res[list(reactive_flexres.keys())[0]]
            ca = None
            cb = None
            for atom in receptor.atoms(idxs):
                if atom["name"] == "CA":
                    ca = atom["xyz"]
                if atom["name"] == "CB":
                    cb = atom["xyz"]
            if ca is None or cb is None:
                check(success=False, error_msg="could not find CA or CB in %s" % reactive_flexres[0])
            v = (cb - ca)
            v /= math.sqrt(v[0]**2 + v[1]**2 + v[2]**2) + 1e-8
            box_center = ca + 5 * v
            box_size = args.box_size
        elif args.ligand is not None:
            box_center, box_size = gridbox.calc_box(args.ligand, args.padding)
        else:
            print("Error: No box center specified.", file=sys.stderr)
            sys.exit(2)

        # write .dat parameter file for B and Si
        ff_fn = pathlib.Path(rigid_fn).parents[0] / pathlib.Path("boron-silicon-atom_par.dat")
        written_files_log["filename"].append(str(ff_fn))
        written_files_log["description"].append("atomic parameters for B and Si (for autogrid)")
        with open(ff_fn, "w") as f:
            f.write(gridbox.boron_silicon_atompar)
        rec_types = set(t for (i, t) in enumerate(receptor.atoms()["atom_type"]) if i not in pdbqt["flex_indices"])
        gpf_string, npts = gridbox.get_gpf_string(box_center, box_size, rigid_fn, rec_types, any_lig_base_types,
                                                    ff_param_fname=ff_fn.name)
        # write GPF
        gpf_fn = pathlib.Path(rigid_fn).with_suffix(".gpf")
        written_files_log["filename"].append(str(gpf_fn))
        written_files_log["description"].append("autogrid input file")
        with open(gpf_fn, "w") as f:
            f.write(gpf_string)

        # write a PDB for the box
        box_fn = str(gpf_fn) + ".pdb"
        written_files_log["filename"].append(box_fn)
        written_files_log["description"].append("PDB file to visualize the grid box")
        with open(box_fn, "w") as f:
            f.write(gridbox.box_to_pdb_string(box_center, npts))

        # check all flexres are inside the box
        any_outside = False
        for atom in receptor.atoms(pdbqt["flex_indices"]):
            if gridbox.is_point_outside_box(atom["xyz"], box_center, npts):
                print("WARNING: Flexible residue outside box." + os_linesep, file=sys.stderr)
                print("WARNING: Strongly recommended to use a box that encompasses flexible residues." + os_linesep, file=sys.stderr)
                break # only need to warn once

    # configuration info for AutoDock-GPU reactive docking
    if len(reactive_flexres) > 0:
        any_lig_reac_types = []
        for order in (1, 2, 3):
            for t in any_lig_base_types:
                any_lig_reac_types.append(reactive_typer.get_reactive_atype(t, order))

        rec_reac_types = []
        for line in all_flex_pdbqt.split(os_linesep):
            if line.startswith("ATOM") or line.startswith("HETATM"):
                atype = line[77:].strip()
                basetype, _ = reactive_typer.get_basetype_and_order(atype)
                if basetype is not None: # is None if not reactive
                    rec_reac_types.append(line[77:].strip())

        derivtypes, modpairs, collisions = get_reactive_config(
                                        any_lig_reac_types,
                                        rec_reac_types,
                                        args.eps_12,
                                        args.r_eq_12,
                                        args.r_eq_13_scaling,
                                        args.r_eq_14_scaling)

        if len(collisions) > 0:
            collision_str = ""
            for t1, t2 in collisions:
                collision_str += "%3s %3s" % (t1, t2) + os_linesep
            collision_fn = str(outpath.with_suffix(".atype_collisions"))
            written_files_log["filename"].append(collision_fn)
            written_files_log["description"].append("type pairs (n=%d) that may lead to intra-molecular reactions" % len(collisions))
            with open(collision_fn, "w") as f:
                f.write(collision_str)

        # The maps block is to tell AutoDock-GPU the base types for the reactive types.
        # This could be done with -T/--derivtypes, but putting derivtypes and intnbp
        # lines in a single configuration file simplifies the command line call.
        map_block = ""
        map_prefix = pathlib.Path(rigid_fn).with_suffix("").name
        all_types = []
        for basetype, reactypes in derivtypes.items():
            all_types.append(basetype)
            map_block += "map %s.%s.map" % (map_prefix, basetype) + os_linesep
            for reactype in reactypes:
                all_types.append(reactype)
                map_block += "map %s.%s.map" % (map_prefix, basetype) + os_linesep
        config = "ligand_types " + " ".join(all_types) + os_linesep
        config += "fld %s.maps.fld" % map_prefix + os_linesep
        config += map_block

        # in modpairs (dict): types are keys, parameters are values
        # now we will write a configuration file with nbp keywords
        # that AD-GPU reads using the --import_dpf flag
        # nbp stands for "non-bonded potential" or "non-bonded pairwise"
        line = "intnbp_r_eps %8.6f %8.6f %3d %3d %4s %4s" + os_linesep
        nbp_count = 0
        for (t1, t2), param in modpairs.items():
            config += line % (param["r_eq"], param["eps"], param["n"], param["m"], t1, t2)
            nbp_count += 1
        config_fn = str(outpath.with_suffix(".reactive_config"))
        written_files_log["filename"].append(config_fn)
        written_files_log["description"].append("reactive parameters for AutoDock-GPU")
        with open(config_fn, "w") as f:
            f.write(config)
        print()
        print("For reactive docking, pass the configuration file to AutoDock-GPU:")
        print("    autodock_gpu -C 1 --import_dpf %s --flexres %s -L <ligand_filename>" % (config_fn, flex_fn))
        print()

    print()
    print("Files written:")
    longest_fn = max([len(fn) for fn in written_files_log["filename"]])
    line = "%%%ds <-- " % longest_fn + "%s"
    for fn, desc in zip(written_files_log["filename"], written_files_log["description"]):
        print(line % (fn, desc))
    return

if __name__ == "__main__":
    sys.exit(main())



from collections import namedtuple

# named tuple to contain information about an atom
PDBAtomInfo = namedtuple('PDBAtomInfo', "name resName resNum chain")
PDBResInfo  = namedtuple('PDBResInfo',       "resName resNum chain")

# standard nucleic acid residue names
nucleic = ['U', 'A', 'C', 'G', 'T']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko analysis
#

from .fingerprint_interactions import FingerprintInteractions
from .interactions import Hydrophobic, Reactive, Metal
from .interactions import HBDonor, HBAcceptor, WaterDonor, WaterAcceptor

__all__ = ['FingerprintInteractions', 
           'Hydrophobic', 'Reactive', 'Metal', 
           'HBDonor', 'HBAcceptor', 'WaterDonor', 'WaterAcceptor']



#!/usr/bin/env python
#
# Combine OA and HD maps for providing the W map
# for water molecules affinity
# The first map should be OA, the second HD
#
# $Header: /mnt/raid/services/cvs/python/packages/share1.5/AutoDockTools/Utilities24/compute_water_map.py,v 1.1 2010/09/30 17:21:36 rhuey Exp $
#
# $Id: compute_water_map.py,v 1.1 2010/09/30 17:21:36 rhuey Exp $
#
# Authors: Stefano Forli, Ruth Huey
#
#

from AutoDockTools.waterMapBuilder import WaterMapBuilder

if __name__ == '__main__':
    import sys
    import getopt


    def usage():
        "Print helpful, accurate usage statement to stdout."
        print """\n
                                        /\/\   __ _ _ __  
                                       /    \ / _` | '_ \ 
                                      / /\/\ \ (_| | |_) |           .-.
                                      \/    \/\__,_| .__/           (   )_
                                                   |_|              _/-'(_)
                                   __    __      _                 (_)
                                  / / /\ \ \__ _| |_ ___ _ __ 
                                  \ \/  \/ / _` | __/ _ \ '__|
                                   \  /\  / (_| | ||  __/ |   
                                    \/  \/ \__,_|\__\___|_|   
                                  
        """
        print "\tINPUT\n\t\tEither the receptor filename or the map files.\n"
        print "\t\t\t-r protein.pdbqt    :   the receptor filename is used to guess the OA and HD filenames"
        print "\n\t\t\t\t *OR*\n"
        print "\t\t\t-o protein.OA.map   :   OA map filename (omit if -r is used)"
        print "\t\t\t-h protein.HD.map   :   HD map filename (omit if -r is used)"
        print "\tOUTPUT\n\t\tW map.\n"
        print "\tOPTIONS\n"
        print "\t\t-m (best,avg,coop..):   mix method (default 'best')"
        print "\t\t                        (Other methods are 'avg','coop', 'boost','best_w')"
        print "\t\t-e float            :   entropy penalty value (default: 0.0)"
        print "\t\t-O float            :   oxygen map weight (default: 1.0)"
        print "\t\t-H float            :   hydrogen map weight (default: 1.0)"
        print "\t\t-s protein.W.map    :   W map output filename" 
        print "\t\t   (default: protein.W.map.[mix method].w[eight].o[xygen weight].h[ydrogen weight].E[ntropy correction])"
        print "\n\n"


    # process command arguments
    try:
        opt_list, args = getopt.getopt(sys.argv[1:], 'o:h:m:e:O:H:s:w:r:vU')
        # -o protein.OA.map
        # -h protein.HD.map
        # -m mix_method best,avg,coop...
        # -e entropy
        # -O oxygen_weight
        # -H hydrogen_boost
        # -s protein.W.map
        # -w weight 0.5
        # -r protein.pdbqt
        # -v verbose
        # -U print Usage statement
    except getopt.GetoptError, msg:
        print 'compute_water_map.py: %s' %msg
        usage()
        sys.exit(2)
        #exit(2) #???

    # initialize required parameters
    # -o protein.OA.map
    oa_map = None
    # -h protein.HD.map
    hd_map = None
    # -m mix_method or mode
    mix_method = 'best'  #best,avg, coop, boost
    # -e entropy 
    entropy = 0.0
    # -O oxygen_weight
    O_weight = 1.0
    # -H hd_boost
    hd_boost = 1.0
    # -s protein.W.map
    w_map = None
    # -w weight
    weight = 0.5
    default_weight = weight
    # -r protein.pdbqt
    protein = None

    ENTROPY = 0.0
    verbose = False

    for o, a in opt_list:
        #print "o=", o, " a=", a
        if o in ('-r', '--r'):
            protein =  os.path.basename(a)
            if verbose: print "set protein to ",a
            name = os.path.splitext(protein)[0]
            #name = protein.rsplit(".")[0]
            oxygenMap = name+".OA.map"
            hydrogenMap = name+".HD.map"
            if verbose:
                print "      OA map ->", oxygenMap
                print "      HD map ->", hydrogenMap
        if o in ('-o', '--o'):
            oa_map = a
        if o in ('-h', '--h'):
            hd_map = a

        if o in ('-w','--w'):
            try:
                default_weight = float(a)
                weight = default_weight
            except:
                print "exception using default_weight", a
                exit(1)
        if o in ('-m', '--m'):
            mix_method = a
            if verbose: print "mix_method=", mix_method
            if mix_method == "BOOST":
                default_weight /= .05

        if o in ("-e", "--e"):
            ENTROPY = float(a)

        if o in ("-O", "--O"):
            O_weight = float(a)

        if o in ("-H", "--H"):
            hd_boost = float(a)


        if o in ("-s","--s"):
            #outputFile = a
            w_map = a
        
        if o in ("-U","--U"):
            usage()
            sys.exit()


    if protein is None and oa_map is None and hd_map is None:        
        usage()
        sys.exit()

    if w_map is None:        
        w_map = "protein.W.map.%s.w%1.2f.O%1.1f.H%1.1f.E%1.1f" % (mix_method, weight, O_weight, hd_boost, ENTROPY )
        print "using default outputfile name: ", w_map
        # protein.W.map.BEST.w0.5.O0.3.H1.3.E0


    # initialize required parameters
    print "\n  MapWater calculator\n ====================="
    print "  mode      :  ", mix_method
    print "  weight    :  ", default_weight
    print "  OA_weight :  ", O_weight
    print "  hd_boost  :  ", hd_boost
    print "  Entropy   :  ", ENTROPY
    print "  output    :  ", w_map


    # optional parameters
    #    if o in ('-h', '--'):
    #        usage()
    #        sys.exit()

    if weight==default_weight and verbose:
        print " => Water map weight : DEFAULT [ %1.2f ]" % weight
    wmb = WaterMapBuilder(OAmap_file=oa_map, HDmap_file=hd_map, Wmap_file=w_map,default_weight=default_weight, weight=weight, mix_method=mix_method)
    wmb.build()
# To execute this command type:
# compute_water_map.py 



from os import linesep as os_linesep
import pathlib
import numpy as np

def get_gpf_string(center, size_angstrom, rec_fname, rec_types, lig_types, map_prefix=None, dielectric=-42, smooth=0.5, spacing=0.375, ff_param_fname=None):

    size_x, size_y, size_z = size_angstrom
    # the following guarantees an EVEN number of grid points
    npts_x = 2 * int(size_x / (2 * spacing))
    npts_y = 2 * int(size_y / (2 * spacing))
    npts_z = 2 * int(size_z / (2 * spacing))

    if map_prefix is None:
        map_prefix = pathlib.Path(rec_fname).with_suffix("").name

    gpf = ( "npts {gpf_npts_x:d} {gpf_npts_y:d} {gpf_npts_z:d}\n"
            "gridfld {map_prefix:s}.maps.fld\n"
            "spacing {gpf_spacing:1.3f}\n"
            "receptor_types {rec_types:s}\n"
            "ligand_types {lig_types_std:s}\n"
            "receptor {rec_fname:s}\n"
            "gridcenter {gpf_gridcenter_x:3.3f} {gpf_gridcenter_y:3.3f} {gpf_gridcenter_z:3.3f}\n"
            "smooth {gpf_smooth:3.3f}\n"
            "{mapfiles_std:s}"
            "elecmap {map_prefix:s}.e.map\n"
            "dsolvmap {map_prefix:s}.d.map\n"
            "dielectric {gpf_dielectric:3.3f}\n")

    if ff_param_fname is not None:
        gpf = "parameter_file %s\n" % ff_param_fname + gpf

    info = {}
    info["gpf_spacing"] = spacing
    info["gpf_smooth"] = smooth
    info["gpf_dielectric"] = dielectric
    info['map_prefix'] = map_prefix
    info['ff_param_fname'] = ff_param_fname
    info['rec_types'] = ' '.join(rec_types)
    info['rec_fname'] = rec_fname
    info['gpf_npts_x'] = npts_x
    info['gpf_npts_y'] = npts_y
    info['gpf_npts_z'] = npts_z
    info['gpf_gridcenter_x'] = center[0]
    info['gpf_gridcenter_y'] = center[1]
    info['gpf_gridcenter_z'] = center[2]

    mapfiles = ""
    for atype in lig_types:
        mapfiles+= "map %s.%s.map\n" % (map_prefix, atype)

    info['lig_types_std'] = " ".join(lig_types)
    info['mapfiles_std'] = mapfiles
    return gpf.format(**info), (npts_x, npts_y, npts_z)

def box_to_pdb_string(box_center, npts, spacing=0.375):
    """
        8 ______ 7
         /.    /|
      4 /_.___/3|
        | . X | |  <-- 9
        |5....|./6
        |.____|/
       1      2

    """

    step_x = int(npts[0] / 2.0) * spacing
    step_y = int(npts[1] / 2.0) * spacing
    step_z = int(npts[2] / 2.0) * spacing
    center_x, center_y, center_z = box_center
    corners = []
    corners.append([center_x - step_x, center_y - step_y, center_z - step_z] ) # 1
    corners.append([center_x + step_x, center_y - step_y, center_z - step_z] ) # 2
    corners.append([center_x + step_x, center_y + step_y, center_z - step_z] ) # 3
    corners.append([center_x - step_x, center_y + step_y, center_z - step_z] ) # 4
    corners.append([center_x - step_x, center_y - step_y, center_z + step_z] ) # 5
    corners.append([center_x + step_x, center_y - step_y, center_z + step_z] ) # 6
    corners.append([center_x + step_x, center_y + step_y, center_z + step_z] ) # 7
    corners.append([center_x - step_x, center_y + step_y, center_z + step_z] ) # 8

    count = 1
    res = "BOX"
    chain = "X"
    pdb_out = ""
    line = "ATOM  %5d %4s %3s %1s%4d    %8.3f%8.3f%8.3f  1.00 10.00          %1s" + os_linesep
    for idx in range(len(corners)):
    	x = corners[idx][0]
    	y = corners[idx][1]
    	z = corners[idx][2]
    	pdb_out += line % (count, "Ne", res, chain, idx+1, x, y, z, "Ne")
    	count += 1

    # center
    pdb_out += line % (count+1, "Xe", res, chain, idx+1, center_x, center_y, center_z, "Xe")

    pdb_out += "CONECT    1    2" + os_linesep
    pdb_out += "CONECT    1    4" + os_linesep
    pdb_out += "CONECT    1    5" + os_linesep
    pdb_out += "CONECT    2    3" + os_linesep
    pdb_out += "CONECT    2    6" + os_linesep
    pdb_out += "CONECT    3    4" + os_linesep
    pdb_out += "CONECT    3    7" + os_linesep
    pdb_out += "CONECT    4    8" + os_linesep
    pdb_out += "CONECT    5    6" + os_linesep
    pdb_out += "CONECT    5    8" + os_linesep
    pdb_out += "CONECT    6    7" + os_linesep
    pdb_out += "CONECT    7    8" + os_linesep
    return pdb_out

def is_point_outside_box(point, center, npts, spacing=0.375):

    # Autogrid always outputs an odd number of grid points, roundup if even.
    # An odd number of grid poins is an even number of voxels.
    halfvoxels = np.array(npts) / 2
    step = spacing * halfvoxels.astype(int)
    mincorner = center - step
    maxcorner = center + step
    x, y, z = point
    is_outside = False
    is_outside |= x >= maxcorner[0] or x <= mincorner[0]
    is_outside |= y >= maxcorner[1] or y <= mincorner[1]
    is_outside |= z >= maxcorner[2] or z <= mincorner[2]
    return is_outside

def calc_box(fname, padding):
    """Crude PDBQT parsing is used here because it allows input of autosite pdbqts"""
    padding = float(padding)
    x_min = float('inf')
    x_min = float('inf')
    y_min = float('inf')
    z_min = float('inf')
    x_max = float('-inf')
    y_max = float('-inf')
    z_max = float('-inf')
    with open(fname) as f:
        for line in f:
            if line.startswith('ATOM') or line.startswith('HETATM'):
                x = float(line[30:38])
                y = float(line[38:46])
                z = float(line[46:54])
                x_max = max(x, x_max)
                y_max = max(y, y_max)
                z_max = max(z, z_max)
                x_min = min(x, x_min)
                y_min = min(y, y_min)
                z_min = min(z, z_min)
    center_x = (x_min + x_max) / 2.0
    center_y = (y_min + y_max) / 2.0
    center_z = (z_min + z_max) / 2.0
    size_x = x_max - x_min + 2 * padding
    size_y = y_max - y_min + 2 * padding
    size_z = z_max - z_min + 2 * padding
    return (center_x, center_y, center_z), (size_x, size_y, size_z)

boron_silicon_atompar  = "atom_par Si     4.10  0.200  35.8235  -0.00143  0.0  0.0  0  -1  -1  6" + os_linesep
boron_silicon_atompar += "atom_par B      3.84  0.155  29.6478  -0.00152  0.0  0.0  0  -1  -1  0" + os_linesep



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko
#


from rdkit import Chem
from rdkit.Geometry import Point3D
from rdkit.Chem import AllChem
from io import StringIO
import json
import os


class RDKitMolCreate:

    ambiguous_flexres_choices = {
        "HIS": ["HIE", "HID", "HIP"],
        "ASP": ["ASP", "ASH"],
        "GLU": ["GLU", "GLH"],
        "CYS": ["CYS", "CYM"],
        "LYS": ["LYS", "LYN"],
        "ARG": ["ARG", "ARG_mgltools"],
        "ASN": ["ASN", "ASN_mgltools"],
        "GLN": ["GLN", "GLN_mgltools"],
    }

    flexres = {
        "CYS": {
            "smiles": "CCS",
            "atom_names_in_smiles_order": ["CA", "CB", "SG"],
            "h_to_parent_index": {"HG": 2},
        },
        "CYM": {
            "smiles": "CC[S-]",
            "atom_names_in_smiles_order": ["CA", "CB", "SG"],
            "h_to_parent_index": {},
        },
        "ASP": {
            "smiles": "CCC(=O)[O-]",
            "atom_names_in_smiles_order": ["CA", "CB", "CG", "OD1", "OD2"],
            "h_to_parent_index": {},
        },
        "ASH": {
            "smiles": "CCC(=O)O",
            "atom_names_in_smiles_order": ["CA", "CB", "CG", "OD1", "OD2"],
            "h_to_parent_index": {"HD2": 4},
        },
        "GLU": {
            "smiles": "CCCC(=O)[O-]",
            "atom_names_in_smiles_order": ["CA", "CB", "CG", "CD", "OE1", "OE2"],
            "h_to_parent_index": {},
        },
        "GLH": {
            "smiles": "CCCC(=O)O",
            "atom_names_in_smiles_order": ["CA", "CB", "CG", "CD", "OE1", "OE2"],
            "h_to_parent_index": {"HE2": 5},
        },
        "PHE": {
            "smiles": "CCc1ccccc1",
            "atom_names_in_smiles_order": ["CA", "CB", "CG", "CD1", "CE1", "CZ", "CE2", "CD2"],
            "h_to_parent_index": {},
        },
        "HIE" : {
            "smiles": "CCc1c[nH]cn1",
            "atom_names_in_smiles_order": ["CA", "CB", "CG", "CD2", "NE2", "CE1", "ND1"],
            "h_to_parent_index": {"HE2": 4},
        },
        "HID" : {
            "smiles": "CCc1cnc[nH]1",
            "atom_names_in_smiles_order": ["CA", "CB", "CG", "CD2", "NE2", "CE1", "ND1"],
            "h_to_parent_index": {"HD1": 6},
        },
        "HIP" : {
            "smiles": "CCc1c[nH+]c[nH]1",
            "atom_names_in_smiles_order": ["CA", "CB", "CG", "CD2", "NE2", "CE1", "ND1"],
            "h_to_parent_index": {"HE2": 4, "HD1": 6},
        },
        "ILE": {
            "smiles": "CC(C)CC",
            "atom_names_in_smiles_order": ["CA", "CB", "CG2", "CG1", "CD1"],
            "h_to_parent_index": {},
        },
        "LYS": {
            "smiles": "CCCCC[NH3+]",
            "atom_names_in_smiles_order": ["CA", "CB", "CG", "CD", "CE", "NZ"],
            "h_to_parent_index": {"HZ1": 5, "HZ2": 5, "HZ3": 5},
        },
        "LYN": {
            "smiles": "CCCCCN",
            "atom_names_in_smiles_order": ["CA", "CB", "CG", "CD", "CE", "NZ"],
            "h_to_parent_index": {"HZ2": 5, "HZ3": 5},
        },
        "LEU": {
            "smiles": "CCC(C)C",
            "atom_names_in_smiles_order": ["CA", "CB", "CG", "CD1", "CD2"],
            "h_to_parent_index": {},
        },
        "MET": {
            "smiles": "CCCSC",
            "atom_names_in_smiles_order": ["CA", "CB", "CG", "SD", "CE"],
            "h_to_parent_index": {},
        },
        "ASN": {
            "smiles": "CCC(=O)N",
            "atom_names_in_smiles_order": ["CA", "CB", "CG", "OD1", "ND2"],
            "h_to_parent_index": {"HD21": 4, "HD22": 4},
        },
        "ASN_mgltools": {
            "smiles": "CCC(=O)N",
            "atom_names_in_smiles_order": ["CA", "CB", "CG", "OD1", "ND2"],
            "h_to_parent_index": {"1HD2": 4, "2HD2": 4},
        },
        "GLN": {
            "smiles": "CCCC(=O)N",
            "atom_names_in_smiles_order": ["CA", "CB", "CG", "CD", "OE1", "NE2"],
            "h_to_parent_index": {"HE21": 5, "HE22": 5},
        },
        "GLN_mgltools": {
            "smiles": "CCCC(=O)N",
            "atom_names_in_smiles_order": ["CA", "CB", "CG", "CD", "OE1", "NE2"],
            "h_to_parent_index": {"1HE2": 5, "2HE2": 5},
        },
        "ARG": {
            "smiles": "CCCCNC(N)=[NH2+]",
            "atom_names_in_smiles_order": ["CA", "CB", "CG", "CD", "NE", "CZ", "NH1", "NH2"],
            "h_to_parent_index": {"HE": 4, "HH11": 6, "HH12": 6, "HH21": 7, "HH22": 7},
        },
        "ARG_mgltools": {
            "smiles": "CCCCNC(N)=[NH2+]",
            "atom_names_in_smiles_order": ["CA", "CB", "CG", "CD", "NE", "CZ", "NH1", "NH2"],
            "h_to_parent_index": {"HE": 4, "1HH1": 6, "2HH1": 6, "1HH2": 7, "2HH2": 7},
        },
        "SER": {
            "smiles": "CCO",
            "atom_names_in_smiles_order": ["CA", "CB", "OG"],
            "h_to_parent_index": {"HG": 2},
        },
        "THR": {
            "smiles": "CC(C)O",
            "atom_names_in_smiles_order": ["CA", "CB", "CG2", "OG1"],
            "h_to_parent_index": {"HG1": 3},
        },
        "VAL": {
            "smiles": "CC(C)C",
            "atom_names_in_smiles_order": ["CA", "CB", "CG1", "CG2"],
            "h_to_parent_index": {},
        },
        "TRP": {
            "smiles": "CCc1c[nH]c2c1cccc2",
            "atom_names_in_smiles_order": ["CA", "CB", "CG", "CD1", "NE1", "CE2", "CD2", "CE3", "CZ3", "CH2", "CZ2"],
            "h_to_parent_index": {"HE1": 4},
        },
        "TYR": {
            "smiles": "CCc1ccc(cc1)O",
            "atom_names_in_smiles_order": ["CA", "CB", "CG", "CD1", "CE1", "CZ", "CE2", "CD2", "OH"],
            "h_to_parent_index": {"HH": 8},
        },
    }

    @classmethod
    def from_pdbqt_mol(cls, pdbqt_mol, only_cluster_leads=False): # TODO add pseudo-water (W atoms, variable nr each pose)
        if only_cluster_leads and len(pdbqt_mol._pose_data["cluster_leads_sorted"]) == 0:
            raise RuntimeError("no cluster_leads in pdbqt_mol but only_cluster_leads=True")
        mol_list = []
        for mol_index in pdbqt_mol._atom_annotations["mol_index"]:
            smiles = pdbqt_mol._pose_data['smiles'][mol_index]
            index_map = pdbqt_mol._pose_data['smiles_index_map'][mol_index]
            h_parent = pdbqt_mol._pose_data['smiles_h_parent'][mol_index]
            atom_idx = pdbqt_mol._atom_annotations["mol_index"][mol_index]

            if smiles is None: # probably a flexible sidechain, but can be another ligand
                residue_names = set()
                atom_names = []
                for atom in pdbqt_mol.atoms(atom_idx):
                    residue_names.add(atom[4])
                    atom_names.append(atom[2])
                if len(residue_names) == 1:
                    resname = residue_names.pop()
                    smiles, index_map, h_parent = cls.guess_flexres_smiles(resname, atom_names)
                    if smiles is None: # failed guessing smiles for possible flexres
                        mol_list.append(None)
                        continue

            if only_cluster_leads:
                pose_ids = pdbqt_mol._pose_data["cluster_leads_sorted"]
            else:
                pose_ids = range(pdbqt_mol._pose_data["n_poses"])

            mol = Chem.MolFromSmiles(smiles)
            coordinates_all_poses = []
            for i in pose_ids:
                pdbqt_mol._current_pose = i
                coordinates = pdbqt_mol.positions(atom_idx)
                mol = cls.add_pose_to_mol(mol, coordinates, index_map) 
                coordinates_all_poses.append(coordinates) 

            # add Hs only after all poses are added as conformers
            # because Chem.AddHs() will affect all conformers at once 
            mol = cls.add_hydrogens(mol, coordinates_all_poses, h_parent) 

            mol_list.append(mol)
        return mol_list

    @classmethod
    def guess_flexres_smiles(cls, resname, atom_names):
        """ Determine a SMILES string for flexres based on atom names,
            as well as the equivalent of smile_index_map and smiles_h_parent
            which are written to PDBQT remarks for regular small molecules.

        Args:
            resname (str):
        
        Returns:
            smiles: SMILES string starting at C-alpha (excludes most of the backbone)
            index_map: list of pairs of integers, first in pair is index in the smiles,
                       second is index of corresponding atom in atom_names         
            h_parent: list of pairs of integers, first in pair is index of a heavy atom
                      in the smiles, second is index of a hydrogen in atom_names.
                      The hydrogen is bonded to the heavy atom. 
        """



        if len(set(atom_names)) != len(atom_names):
            return None, None, None
        candidate_resnames = cls.ambiguous_flexres_choices.get(resname, [resname])
        for resname in candidate_resnames:
            is_match = False
            if resname not in cls.flexres[resname]["atom_names_in_smiles_order"]:
                continue
            atom_names_in_smiles_order = cls.flexres[resname]["atom_names_in_smiles_order"]
            h_to_parent_index = cls.flexres[resname]["h_to_parent_index"]
            expected_names = atom_names_in_smiles_order + list(h_to_parent_index.keys())
            if len(atom_names) != len(expected_names):
                continue
            nr_matched_atom_names = sum([int(n in atom_names) for n in expected_names])
            if nr_matched_atom_names == len(expected_names):
                is_match = True
                break
        if not is_match:
            return None, None, None
        else:
            smiles = cls.flexres[resname]["smiles"]
            index_map = []
            for smiles_index, name in enumerate(atom_names_in_smiles_order):
                index_map.append(smiles_index + 1) 
                index_map.append(atom_names.index(name) + 1)
            h_parent = []
            for name, smiles_index in h_to_parent_index.items():
                h_parent.append(smiles_index + 1)
                h_parent.append(atom_names.index(name) + 1)
            return smiles, index_map, h_parent

    @classmethod
    def add_pose_to_mol(cls, mol, ligand_coordinates, index_map):
        """add given coordinates to given molecule as new conformer.
        Index_map maps order of coordinates to order in smile string
        used to generate rdkit mol

        Args:
            ligand_coordinates: 2D array of shape (nr_atom, 3).
            index_map: list of nr_atom pairs of integers, 1-indexed.
                       In each pair, the first int is the index in mol, and
                       the second int is the index in ligand_coordinates

        Raises:
            RuntimeError: Will raise error if number of coordinates provided does not
                match the number of atoms there should be coordinates for.
        """

        n_atoms = mol.GetNumAtoms()
        n_mappings = int(len(index_map) / 2)
        conf = Chem.Conformer(n_atoms)
        if n_atoms < n_mappings:
            raise RuntimeError(
                "Given {n_coords} atom coordinates "
                "but index_map is greater at {n_at} atoms.".format(
                    n_coords=n_atoms, n_at=n_mappings))
        coord_is_set = [False] * n_atoms
        for i in range(n_mappings):
            pdbqt_index = int(index_map[i * 2 + 1]) - 1
            mol_index = int(index_map[i * 2]) - 1
            x, y, z = [float(coord) for coord in ligand_coordinates[pdbqt_index]]
            conf.SetAtomPosition(mol_index, Point3D(x, y, z))
            coord_is_set[mol_index] = True
        mol.AddConformer(conf, assignId=True)
        # some hydrogens (isotopes) may have no coordinate set yet
        for i, is_set in enumerate(coord_is_set):
            if not is_set:
                atom = mol.GetAtomWithIdx(i)
                if atom.GetAtomicNum() != 1:
                    raise RuntimeError("Only H allowed to be in SMILES but not in PDBQT")
                neigh = atom.GetNeighbors()
                if len(neigh) != 1:
                    raise RuntimeError("Expected H to have one neighbor")
                AllChem.SetTerminalAtomCoords(mol, i, neigh[0].GetIdx())
        return mol


    @staticmethod
    def add_hydrogens(mol, coordinates_list, h_parent):
        """Add hydrogen atoms to ligand RDKit mol, adjust the positions of
            polar hydrogens to match pdbqt
        """
        mol = Chem.AddHs(mol, addCoords=True)
        conformers = list(mol.GetConformers())
        num_hydrogens = int(len(h_parent) / 2)
        for conformer_idx, atom_coordinates in enumerate(coordinates_list):
            conf = conformers[conformer_idx]
            used_h = []
            for i in range(num_hydrogens):
                parent_rdkit_index = h_parent[2 * i] - 1
                h_pdbqt_index = h_parent[2 * i + 1] - 1
                x, y, z = [
                    float(coord) for coord in atom_coordinates[h_pdbqt_index]
                ]
                parent_atom = mol.GetAtomWithIdx(parent_rdkit_index)
                candidate_hydrogens = [
                    atom.GetIdx() for atom in parent_atom.GetNeighbors()
                    if atom.GetAtomicNum() == 1
                ]
                for h_rdkit_index in candidate_hydrogens:
                    if h_rdkit_index not in used_h:
                        break
                used_h.append(h_rdkit_index)
                conf.SetAtomPosition(h_rdkit_index, Point3D(x, y, z))
        return mol

    @staticmethod
    def combine_rdkit_mols(mol_list):
        """Combines list of rdkit molecules into a single one
            None's are ignored
            returns None if input is empty list or all molecules are None
        """
        combined_mol = None
        for mol in mol_list:
            if mol is None:
                continue
            if combined_mol is None: # first iteration
                combined_mol = mol
            else:
                combined_mol = Chem.CombineMols(combined_mol, mol)
        return combined_mol

    @classmethod
    def _verify_flexres(cls):
        for resname in cls.flexres:
            atom_names_in_smiles_order = cls.flexres[resname]["atom_names_in_smiles_order"]
            h_to_parent_index = cls.flexres[resname]["h_to_parent_index"]
            expected_names = atom_names_in_smiles_order + list(h_to_parent_index.keys())
            if len(expected_names) != len(set(expected_names)):
                raise RuntimeError("repeated atom names in cls.flexres[%s]" % resname)

    @staticmethod
    def write_sd_string(pdbqt_mol, only_cluster_leads=False):
        sio = StringIO()
        f = Chem.SDWriter(sio)
        mol_list = RDKitMolCreate.from_pdbqt_mol(pdbqt_mol, only_cluster_leads)
        failures = [i for i, mol in enumerate(mol_list) if mol is None]
        combined_mol = RDKitMolCreate.combine_rdkit_mols(mol_list)
        if combined_mol is None:
            return "", failures
        nr_conformers = combined_mol.GetNumConformers()
        property_names = {
            "free_energy": "free_energies",
            "intermolecular_energy": "intermolecular_energies",
            "internal_energy": "internal_energies",
            "cluster_size": "cluster_size",
            "cluster_id": "cluster_id",
            "rank_in_cluster": "rank_in_cluster",
        }
        props = {}
        if only_cluster_leads:
            nr_poses = len(pdbqt_mol._pose_data["cluster_leads_sorted"])
            pose_idxs = pdbqt_mol._pose_data["cluster_leads_sorted"]
        else:
            nr_poses = pdbqt_mol._pose_data["n_poses"]
            pose_idxs = list(range(nr_poses))
        for prop_sdf, prop_pdbqt in property_names.items():
            if nr_conformers == nr_poses:
                props[prop_sdf] = prop_pdbqt
        has_all_data = True
        for _, key in props.items():
            has_all_data &= len(pdbqt_mol._pose_data[key]) == nr_conformers
        for conformer in combined_mol.GetConformers():
            i = conformer.GetId()
            j = pose_idxs[i]
            if has_all_data:
                data = {k: pdbqt_mol._pose_data[v][j] for k, v in props.items()}
                if len(data): combined_mol.SetProp("meeko", json.dumps(data))
            f.write(combined_mol, i)
        f.close()
        output_string = sio.getvalue()
        return output_string, failures



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko
#

import os

import numpy as np
import pandas as pd

from .interactions import Hydrophobic, Reactive, Metal
from .interactions import HBDonor, HBAcceptor, WaterDonor, WaterAcceptor


class FingerprintInteractions:

    def __init__(self, receptor):
        """FingerprintInteractions object

        Args:
            receptor (PDBQTReceptor): receptor 

        """
        self._data = []
        self._interactions = [Hydrophobic(),
                              Reactive(),
                              Metal(), 
                              HBDonor(), HBAcceptor(),
                              WaterDonor(), WaterAcceptor()]
        self._unique_interactions = {interaction.name: {*()} for interaction in self._interactions}
        self._receptor = receptor

    def run(self, molecules):
        """Run the fingerprint interactions.
        
        Args:
            molecules (PDBQTMolecule, list of PDBQTMolecule): molecule or list of molecules

        """
        data = []

        if not isinstance(molecules, (list, tuple)):
            molecules = [molecules]

        for molecule in molecules:
            for pose in molecule:
                tmp = {}

                for interaction in self._interactions:
                    resids = ''
                    columns = ['chain', 'resid']
                    if interaction.name in ['HBDonor', 'HBAcceptor', 'WaterDonor', 'WaterAcceptor']:
                        columns += ['name']

                    rigid_interactions, flex_interactions = interaction.find(pose, self._receptor)

                    rec_rigid_atoms = self._receptor.atoms(rigid_interactions['receptor_idx'])
                    rec_flex_atoms = pose.atoms(flex_interactions['receptor_idx'])

                    if rec_rigid_atoms.size > 0:
                        unique_resids = np.unique(rec_rigid_atoms[columns])
                        resids = [':'.join([interaction.name] + [str(v) for v in u]) for u in unique_resids]

                    if rec_flex_atoms.size > 0:
                        unique_resids = np.unique(rec_flex_atoms[columns])
                        resids += [':'.join([interaction.name] + [str(v) for v in u]) for u in unique_resids]

                    tmp[interaction.name] = resids

                tmp_data = [pose.name, pose.pose_id]
                for inte_type, resids in tmp.items():
                    unique_resids = set(resids)
                    self._unique_interactions[inte_type].update(unique_resids)
                    tmp_data.append(list(unique_resids))
                data.append(tmp_data)

        self._data.extend(data)

    def to_dataframe(self, remove_common=False):
        """Generate a panda DataFrame with all the interactions
        
        Args:
            remove_common (bool): remove all the interactions (columns) that
                are common to all the molecules (default: False)

        Returns:
            pd.DataFrame: pandas DataFrame containing all the interactions
                found between the molecules and the receptor

        """
        count = 0
        resid_to_idx_encoder = {}
        columns = [[], []]
        names = []
        poses = []

        # Generate one-hot encoder-like (and the columns for the dataframe)
        for inte_type, resids in self._unique_interactions.items():
            columns[0].extend([inte_type] * len(resids))
            # Remove the v/w/h_ tag for the column names
            columns[1].extend([resid.split(':', 1)[-1] for resid in resids])

            for resid in resids:
                resid_to_idx_encoder[resid] = count
                count += 1

        # Create multicolumns for the dataframe
        c_tuples = list(zip(*columns))
        multi_columns = pd.MultiIndex.from_tuples(c_tuples)

        # Convert resids in one hot fingerprint interactions
        fpi = np.zeros(shape=(len(self._data), count), dtype=int)

        for i, pose_molecule in enumerate(self._data):
            resids = [j for i in pose_molecule[2:] for j in i]
            idx = [resid_to_idx_encoder[resid] for resid in resids]
            fpi[i][idx] = 1
            names.append(pose_molecule[0])
            poses.append(pose_molecule[1] + 1)

        # Create final dataframe
        df = pd.DataFrame(fpi, index=np.arange(0, len(self._data)), columns=multi_columns)
        
        df['name'] = names
        df['pose'] = poses
        df.set_index(['name', 'pose'], inplace=True)

        if remove_common:
            # Remove interactions (columns) that are common to all molecules
            df = df.loc[:, (df.sum(axis=0) != df.shape[0])]

        return df



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Reactive
#

from math import sqrt
from os import linesep as os_linesep


class ReactiveAtomTyper:

    def __init__(self):

        self.ff = {
            "HD": {"rii": 2.00, "epsii": 0.020},
            "C":  {"rii": 4.00, "epsii": 0.150},
            "A":  {"rii": 4.00, "epsii": 0.150},
            "N":  {"rii": 3.50, "epsii": 0.160},
            "NA": {"rii": 3.50, "epsii": 0.160},
            "OA": {"rii": 3.20, "epsii": 0.200},
            "OS": {"rii": 3.20, "epsii": 0.200},
            "F":  {"rii": 3.09, "epsii": 0.080},
            "P":  {"rii": 4.20, "epsii": 0.200},
            "SA": {"rii": 4.00, "epsii": 0.200},
            "S":  {"rii": 4.00, "epsii": 0.200},
            "Cl": {"rii": 4.09, "epsii": 0.276},
            "CL": {"rii": 4.09, "epsii": 0.276},
            "Br": {"rii": 4.33, "epsii": 0.389},
            "BR": {"rii": 4.33, "epsii": 0.389},
            "I":  {"rii": 4.72, "epsii": 0.550},
            "Si": {"rii": 4.10, "epsii": 0.200},
            "B":  {"rii": 3.84, "epsii": 0.155},
            "W":  {"rii": 0.00, "epsii": 0.000},
        }
        std_atypes = list(self.ff.keys())
        rt, r2s, r2o = self.enumerate_reactive_types(std_atypes)
        self.reactive_type = rt
        self.reactive_to_std_atype_mapping = r2s
        self.reactive_to_order = r2o

    @staticmethod
    def enumerate_reactive_types(atypes):
        reactive_type = {1:{}, 2:{}, 3:{}}
        reactive_to_std_atype_mapping = {}
        reactive_to_order = {}
        for order in (1,2,3):
            for atype in atypes:
                if len(atype) == 1:
                    new_atype = "%s%d" % (atype, order)
                else:
                    new_atype = "%s%d" % (atype[0], order+3)
                    if new_atype in reactive_to_std_atype_mapping:
                        new_atype = "%s%d" % (atype[0], order+6)
                        if new_atype in reactive_to_std_atype_mapping:
                            raise RuntimeError("ran out of numbers for reactive types :(")
                reactive_to_std_atype_mapping[new_atype] = atype
                reactive_to_order[new_atype] = order
                reactive_type[order][atype] = new_atype
                ### # avoid atom type clashes with multiple reactive residues by
                ### # prefixing with the index of the residue, e.g. C3 -> 1C3.
                ### for i in range(8): # hopefully 8 reactive residues is sufficient
                ###     prefixed_new_atype = '%d%s' % ((i+1), new_atype)
                ###     reactive_to_std_atype_mapping[prefixed_new_atype] = atype
        return reactive_type, reactive_to_std_atype_mapping, reactive_to_order


    def get_scaled_parm(self, atype1, atype2):
        """ generate scaled parameters for a pairwise interaction between two atoms involved in a
            reactive interaction

            Rij and epsij are calculated according to the AD force field:
                # - To obtain the Rij value for non H-bonding atoms, calculate the
                #        arithmetic mean of the Rii values for the two atom types.
                #        Rij = (Rii + Rjj) / 2
                #
                # - To obtain the epsij value for non H-bonding atoms, calculate the
                #        geometric mean of the epsii values for the two atom types.
                #        epsij = sqrt( epsii * epsjj )
        """

        atype1_org, _ = self.get_basetype_and_order(atype1)
        atype2_org, _ = self.get_basetype_and_order(atype2)
        atype1_rii = self.ff[atype1_org]['rii']
        atype1_epsii = self.ff[atype1_org]['epsii']
        atype2_rii = self.ff[atype2_org]['rii']
        atype2_epsii = self.ff[atype2_org]['epsii']
        atype1_rii = atype1_rii
        atype2_rii = atype2_rii
        rij = (atype1_rii + atype2_rii) / 2
        epsij = sqrt(atype1_epsii * atype2_epsii)
        return rij, epsij


    def get_reactive_atype(self, atype, reactive_order):
        """ create or retrive new reactive atom type label name"""
        macrocycle_glue_types = ["CG%d" % i for i in range(9)]
        macrocycle_glue_types += ["G%d" % i for i in range(9)]
        if atype in macrocycle_glue_types:
            return None
        return self.reactive_type[reactive_order][atype]


    def get_basetype_and_order(self, atype):
        if len(atype) > 1:
            if atype[0].isdecimal():
                atype = atype[1:] # reactive residues are prefixed with a digit
        if atype not in self.reactive_to_std_atype_mapping:
            return None, None
        basetype = self.reactive_to_std_atype_mapping[atype]
        order = self.reactive_to_order[atype]
        return basetype, order


reactive_typer = ReactiveAtomTyper()
get_reactive_atype = reactive_typer.get_reactive_atype

def assign_reactive_types(molsetup, smarts, smarts_idx, get_reactive_atype=get_reactive_atype):

    atype_dicts = []
    for atom_indices in molsetup.find_pattern(smarts):
        atypes = molsetup.atom_type.copy()
        reactive_atom_index = atom_indices[smarts_idx]

        # type reactive atom
        original_type = molsetup.atom_type[reactive_atom_index]
        reactive_type = get_reactive_atype(original_type, reactive_order=1) 
        atypes[reactive_atom_index] = reactive_type

        # type atoms 1 bond away from reactive atom
        for index1 in molsetup.graph[reactive_atom_index]:
            if not molsetup.atom_ignore[index1]:
                original_type = molsetup.atom_type[index1]
                reactive_type = get_reactive_atype(original_type, reactive_order=2)
                atypes[index1] = reactive_type

            # type atoms 2 bonds away from reactive
            for index2 in molsetup.graph[index1]:
                if index2 == reactive_atom_index:
                    continue
                if not molsetup.atom_ignore[index2]:
                    original_type = molsetup.atom_type[index2]
                    reactive_type = get_reactive_atype(original_type, reactive_order=3)
                    atypes[index2] = reactive_type

        atype_dicts.append(atypes)

    return atype_dicts


# enumerate atom type pair combinations for reactive docking configuration file

def get_reactive_config(types_1, types_2, eps12, r12, r13_scaling, r14_scaling, ignore=['HD', 'F'], coeff_vdw=.1662):
    """
    Args:
        types_1 (list): 1st set of atom types
        types_2 (list): 2nd set of atom types

    Returns:
        derivtypes (dict):
        modpairs (list):
    """

    # for reactive pair (1-2 interaction) use 13-7 potential
    n12 = 13
    m12 = 7

    # for 1-3 and 1-4 interactions stick to 12-6 potential
    n = 12
    m = 6

    derivtypes = {}
    for reactype in set(types_1).union(set(types_2)):
        basetype, order = reactive_typer.get_basetype_and_order(reactype)
        derivtypes.setdefault(basetype, [])
        derivtypes[basetype].append(reactype)

    scaling = {3: r13_scaling, 4: r14_scaling}

    modpairs = {}
    for t1 in set(types_1):
        basetype_1, order_1 = reactive_typer.get_basetype_and_order(t1)
        for t2 in set(types_2):
            basetype_2, order_2 = reactive_typer.get_basetype_and_order(t2)
            order = order_1 + order_2
            pair_id = tuple(sorted([t1, t2]))
            if order == 2: # 1-2 interaction: these are the reactive atoms.
                modpairs[pair_id] = {"eps": eps12, "r_eq": r12, "n": n12, "m": m12}
            elif order == 3 or order == 4:
                if basetype_1 in ignore or basetype_2 in ignore:
                    rij = 0.01
                    epsij = 0.001
                else:
                    rij, epsij = reactive_typer.get_scaled_parm(t1, t2)
                    rij *= scaling[order]
                    epsij *= coeff_vdw
                modpairs[pair_id] = {"eps": epsij, "r_eq": rij, "n": n, "m": m}

    # pairs of types across sets that also happen within each set
    def enum_pairs(types):
        pairs = set()
        for i in range(len(types_1)):
            for j in range(i, len(types_1)):
                pair_id = tuple(sorted([types_1[i], types_1[j]]))
                pairs.add(pair_id)
        return pairs
    collisions = []
    collisions.extend([p for p in enum_pairs(types_1) if p in modpairs])
    collisions.extend([p for p in enum_pairs(types_2) if p in modpairs])
    collisions = set(collisions)

    return derivtypes, modpairs, collisions



############################################################################
#
# Authors: Stefano Forli, Ruth Huey
#
# Copyright: M. Sanner TSRI 2010
#
#############################################################################

#
# $Header: /opt/cvs/python/packages/share1.5/AutoDockTools/waterMapBuilder.py,v 1.2 2010/09/30 17:13:36 rhuey Exp $
#
# $Id: waterMapBuilder.py,v 1.2 2010/09/30 17:13:36 rhuey Exp $
#

"""
This module implements the WaterMapBuilder class which combines an AutoGrid oxygen map ("receptor.OA.map") 
and an AutoGrid hydrogen affinity map ("receptor.HD.map") to create a water affinity map ("receptor.W.map").

Options and default values include:
default_weight   0.5
mode            "BEST"
entropy          0.0
O_weight         1.0
hd_boost         1.0

"""

class WaterMapBuilder:
    """Class for creating an AutoGrid water map 'stem.W.map' by combining an oxygen map 'stem.OA.map' and a
    hydrogen map 'stem.HD.map'. 
    The available methods for combining the maps based on each pair of values from OA.map and  HD.map include: 
        'best': use the smaller of [OA.map value *self.O_weight, HD.map value *self.hd_boost]
        'avg': if not positive, if either is > use self.ENTROPY =0.0 by default 
               else (OA.map value + HD.map value)/2 ) * self.weight
        'coop': if not positive, if either is > use self.ENTROPY =0.0 by default 
               else ((OA.map value  + HD.map value)) * self.weight
        'boost': if not positive, if either is > use self.ENTROPY =0.0 by default 
                else ((OA.map value *7 + HD.map value *7)/2) * self.weight
        'best_w': if not positive, if either is > use self.ENTROPY =0.0 by default 
                else if OA.map.value < HD.map value :
                    self.O += 1
                    return OA.map.value*self.weight
                else:
                    self.H += 1
                    return HD.map.value*self.weight

    """

    def __init__(self, receptor_file=None, OAmap_file=None, HDmap_file=None, 
                    Wmap_file=None, default_weight=0.5, mix_method="best",
                    ENTROPY=0.0, O_weight=1, hd_boost=1, weight=None, verbose=False):
        #NB: mix_method==mode
        self.receptor_file = receptor_file
        self.OAmap_file = OAmap_file   #firstMap
        self.HDmap_file = HDmap_file   #secondMap
        self.Wmap_file = Wmap_file
        self.default_weight = default_weight
        self.mix_method = mix_method
        self.ENTROPY = ENTROPY
        self.O_weight = O_weight
        self.hd_boost = hd_boost
        self.weight = weight
        if weight is None: 
            weight = default_weight
            self.weight = default_weight
        self.verbose = verbose
        output_stem = None
        if receptor_file is None and (OAmap_file is None or HDmap_file is None):
            raise "receptor filename or both oxygen map file AND hydrogen map file must be specified"
        #???
        self.O = 0
        self.H = 0
        if receptor_file is not None:
            output_stem = os.path.splitext(os.path.basename(receptor_file))[0]
            if self.verbose: print("set output_stem from receptor_file to ", output_stem)
            self.OAmap_file = output_stem + ".OA.map"
            self.HDmap_file = output_stem + ".HD.map"
        elif OAmap_file is not None and HDmap_file is not None:
            output_stem = OAmap_file.split('.')[0] 
            if self.verbose: print("set output_stem from OAmap to ", output_stem)
        if output_stem is None:
            raise "output_stem is None!"
        self.output_stem = output_stem
        if self.Wmap_file is None:
            self.Wmap_file = "protein.W.map.%s.w%1.2f.O%1.1f.H%1.1f.E%1.1f" % (mix_method, weight, O_weight, hd_boost, ENTROPY )

            #by default self.Wmap_file = output_stem + ".W.map"
            if self.verbose: print("set self.Wmap_file to ", self.Wmap_file)
        if self.OAmap_file is None:
            raise "oxygen map file not specified"
        if self.HDmap_file is None:
            raise "hydrogen map file not specified"
        if mix_method not in ['best','avg','coop','boost','best_w']: #mode
            msg = 'Unrecognized mix_method:' + mix_method 
            raise  msg
        if mix_method == 'best':
            self.method = self.best
        elif mix_method == 'avg':
            self.method = self.avg
        elif mix_method == 'coop':
            self.method = self.coop
        elif mix_method == 'boost':
            self.default_weight /= 0.5
            self.method = self.boost
        elif mix_method == 'best_w':
            self.method = self.best_w


    def best(self, first, second, positive=False ): #, hd_boost = hd_boost):
        #global O, H   #@@@@@@@@@@@@@
        """
        - Return the best value between two map values
        - If any positive values are found, 0 is returned by default
        - by default it should be first=OA, second=HD
        """
        #print "in best" 
        first *= self.O_weight
        second *= self.hd_boost
        #print "first, second:", first, second

        if not positive and (first > 0 or second > 0):
                # TODO add a smoothing function for positive values?
                return self.ENTROPY
        else:
            if first < second :
                self.O += 1
                return first * self.weight
            else:
                self.H += 1
                #print "First: %2.6f\tSecond: %2.6f" % (first, second)
                return second * self.weight


    def best_w(self, first, second, positive=False):
        #global O, H   #@@@@@@@@@@@@@
        # TODO test different parameters to check if there's
        # a real advantage
        """
        - Return the best value between two map values
        - If any positive values are found, 0 is returned by default
        - WEIGHTED: the OA map is scaled by weight to account for the higher
                    intensity of OA maps versus HD
        """
        
        first *= self.O_weight
        second *= self.hd_boost

        if not positive and (first > 0 or second > 0):
                return 0 + self.ENTROPY
        else:
            if first < second :
                self.O += 1
                return first*self.weight
            else:
                self.H += 1
                return second*self.weight



    def avg(self, first, second, positive = False):
        """
        - Return the average value between two map values
        - If any positive values are found, 0 is returned by default
        """
        if not positive and (first > 0 or second > 0):
                return self.ENTROPY
        else:
            value = ( (first + second)/2 ) * self.weight
            return value


    def coop(self, first, second, positive = False):
        """
        - Return the sum of value between two map values
        - If any positive values are found, 0 is returned by default
        """
        if not positive and (first > 0 or second > 0):
            return self.ENTROPY
        else:
            value = ((first + second)) * self.weight
            return value


    def boost(self, first, second, positive = False):
        if not positive and (first > 0 or second > 0):
            return self.ENTROPY
        else:
            value =  ((first*7+second*7)/2) * self.weight
            return value


    def build(self):
        if self.verbose:
            print("in build:", end=' ')
            print(" self.OAmap_file=", self.OAmap_file)
            print(" self.HDmap_file=", self.HDmap_file)
            print(" self.Wmap_file=", self.Wmap_file)
        try:
            #MAP1 = open(firstMap, 'r')
            oa_ptr = open(self.OAmap_file)
            oa_lines = oa_ptr.readlines()
            oa_ptr.close()
            if self.verbose: print("read %d lines from %s"%(len(oa_lines),self.OAmap_file))
        except:
            print("ERROR: unable to open map file: %s" % self.OAmap_file)
            exit(1)

        try:
            #MAP2 = open(secondMap, 'r' )
            hd_ptr = open(self.HDmap_file)
            hd_lines = hd_ptr.readlines()
            hd_ptr.close()
            if self.verbose: print("read %d lines from %s"%(len(hd_lines),self.HDmap_file))
        except:
            #print "ERROR: the map %s can't be open" % self.HDmap_file
            print("ERROR: unable to open map file: %s" % self.HDmap_file)
            exit(1)

        #sanity checks:
        assert len(oa_lines)==len(hd_lines)

        try:
            if self.verbose: print("opening ", self.Wmap_file)
            optr = open(self.Wmap_file, 'w' )
            if self.verbose: print("opened water mapfile: ", self.Wmap_file)
            #DIFFERENCE = open(self.Wmap_file, 'w' )
        except:
            print("ERROR: impossible to open the output map", self.Wmap_file)
            exit(1)

        #get the header lines from the oxygen map
        #HEADER = lines[:6]
        for ll in oa_lines[:6]:
            optr.write(ll)
            #if self.verbose: print "wrote ", ll
        # process the rest of the lines
        mode = self.mix_method
        if self.verbose: print("using mode =", mode, 'to create W map_file=', self.Wmap_file)
        for i in range(6, len(hd_lines)):
            oxy_val=oa_lines[i].strip()
            oxy_val = float(oxy_val)
            hyd_val=hd_lines[i].strip()
            hyd_val = float(hyd_val)
            optr.write("%1.4f\n"%self.method(*(oxy_val, hyd_val)))
        optr.close()     
        if self.verbose: print("closed ", self.Wmap_file)

#        # output the results
#        for index in range(len(HEADER)):
#            DIFFERENCE.write(str(HEADER[index]))
#####
#        for index in range(len(total)):
#            try:
#                value = "%1.4f" % (total[index])
#            except:
#                print "problem"
#            DIFFERENCE.write(value+'\n')
#        DIFFERENCE.close
        if mode in ['best', 'best_w']:
            O = float(self.O)
            H = float(self.H)
            tot = H+O
            if tot == 0: raise 'O+H == zero !'
            o_pc = (O/tot)*100
            h_pc = (H/tot)*100

            print("\n     output results   ")
            print("  --------------------")
            print("  OA points : %3.2f%s" % ( o_pc , "%"))
            print("  HD points : %3.2f%s\n" % ( h_pc, "%"))




#!/usr/bin/env python

from rdkit import Chem
from rdkit.Chem import rdDistGeom
from meeko import MoleculePreparation

def test():
    smiles = "C1CCCC1"
    mol = Chem.MolFromSmiles(smiles)
    mol = Chem.AddHs(mol)
    etkdgv3 = rdDistGeom.srETKDGv3()
    rdDistGeom.EmbedMolecule(mol, etkdgv3)
    mk_prep = MoleculePreparation()
    molsetup = mk_prep.prepare(mol)[0]
    nr_rot_bonds = 0
    for idxs, bond_info in molsetup.bond.items():
        nr_rot_bonds += int(bond_info["rotatable"])
    assert(nr_rot_bonds == 0)



from meeko import MoleculePreparation
from meeko import PDBQTMolecule
from meeko import PDBQTWriterLegacy
from rdkit import Chem
import pathlib

num_cycle_breaks = {
    "tetrahedron1": 3,
    "tetrahedron2": 3,
    "vancomycin":   3,
    "macrocycle2":   1,
    "macrocycle3":   1,
    "macrocycle4":   2,
    "macrocycle5":   1,
}
workdir = pathlib.Path(__file__).parents[0]
filenames = {name: str(workdir / "macrocycle_data" / ("%s.sdf" % name)) for name in num_cycle_breaks}
mk_prep = MoleculePreparation()

def get_macrocycle_atom_types(pdbqt_string):
    macrocycle_carbon = ['CG0', 'CG1', 'CG2', 'CG3', 'CG4', 'CG5', 'CG6', 'CG7', 'CG8', 'CG9']
    macrocycle_pseudo = [ 'G0',  'G1',  'G2',  'G3',  'G4',  'G5',  'G6',  'G7',  'G8',  'G9']
    cg_atoms = []
    g_atoms = []
    lines = pdbqt_string.split('\n')
    for line in lines:
        if line.startswith("ATOM") or line.startswith("HETATM"):
            atom_type = line[77:].strip()
            if atom_type in macrocycle_carbon:
                cg_atoms.append(atom_type)
            elif atom_type in macrocycle_pseudo:
                g_atoms.append(atom_type)
    return cg_atoms, g_atoms

def test_external_ring_closure():
    mol_fn = workdir / "macrocycle_data" /"open-ring-3D-graph-intact_small.mol"
    mol = Chem.MolFromMolFile(str(mol_fn), removeHs=False)
    delete_bonds = [(2, 3)]
    glue_pseudos = {2: (-999.9, 0, 0), 3: (42.0, 0, 0.)}
    setups = mk_prep.prepare(
                    mol,
                    delete_ring_bonds=delete_bonds,
                    glue_pseudo_atoms=glue_pseudos)
    assert(len(setups) == 1)
    pdbqt_string, is_ok, error_msg = PDBQTWriterLegacy.write_string(setups[0])
    assert(setups[0].ring_closure_info["bonds_removed"] == [(2, 3)])
    assert(2 in setups[0].ring_closure_info["pseudos_by_atom"])
    assert(3 in setups[0].ring_closure_info["pseudos_by_atom"])
    cg_atoms, g_atoms = get_macrocycle_atom_types(pdbqt_string)
    assert(len(cg_atoms) == len(g_atoms))
    assert(len(cg_atoms) == 2 * len(set(cg_atoms)))
    assert(len(g_atoms) == 2 * len(set(g_atoms)))
    assert(len(set(g_atoms)) == 1)
    p = PDBQTMolecule(pdbqt_string)
    glue_x_coords = []
    for atom in p.atoms():
        if atom["atom_type"].startswith("G"):
            glue_x_coords.append(atom["xyz"][0])
    for atom_index, (x, y, z) in glue_pseudos.items():
        mindist = 99999.9
        for xcoord in glue_x_coords:
            mindist = min(mindist, abs(x - xcoord))
        assert(mindist < 1e-3) # loose matching

def run(molname):
    filename = filenames[molname]
    mol = Chem.MolFromMolFile(filename, removeHs=False)
    setups = mk_prep.prepare(mol)
    pdbqt_string, is_ok, error_msg = PDBQTWriterLegacy.write_string(setups[0])
    cg_atoms, g_atoms = get_macrocycle_atom_types(pdbqt_string)
    assert(len(cg_atoms) == len(g_atoms))
    assert(len(cg_atoms) == 2 * len(set(cg_atoms)))
    assert(len(g_atoms) == 2 * len(set(g_atoms)))
    assert(len(set(g_atoms)) == num_cycle_breaks[molname])

def test_all():
    for molname in num_cycle_breaks:
        run(molname)



from meeko import MoleculePreparation
from meeko import PDBQTWriterLegacy
from rdkit import Chem
from rdkit.Chem import rdDistGeom
import warnings
import pytest


def test():
    p = Chem.SmilesParserParams()
    mol = Chem.MolFromSmiles("Oc1ccc(cc1)[N+](=O)[O-]\tnitrophenol", p)
    etkdg_params = Chem.rdDistGeom.ETKDGv3()
    Chem.rdDistGeom.EmbedMolecule(mol, etkdg_params)
    mk_prep = MoleculePreparation()
    with warnings.catch_warnings(record=True) as cought:
        setups = mk_prep.prepare(mol)
        for w in cought:
            print(w)
    pdbqt, is_ok, error_msg = PDBQTWriterLegacy.write_string(setups[0])
    assert(is_ok == False)

mk_prep = MoleculePreparation()

def test_no_conformer():
    mol = Chem.MolFromSmiles("C1CCCOC1")
    mol = Chem.AddHs(mol)
    with pytest.raises(ValueError) as e:
        mk_prep.prepare(mol)
    assert(str(e.value).endswith("Need 3D coordinates."))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko PDBQT writer
#

import sys
import json
import math

import numpy as np
from rdkit import Chem
from .utils import pdbutils
from .utils.rdkitutils import mini_periodic_table


class PDBQTWriterLegacy():

    @staticmethod
    def _get_pdbinfo_fitting_pdb_chars(pdbinfo):
        """ return strings and integers that are guaranteed
            to fit within the designated chars of the PDB format """

        atom_name = pdbinfo.name
        res_name = pdbinfo.resName
        res_num = pdbinfo.resNum
        chain = pdbinfo.chain
        if len(atom_name) > 4: atom_name = atom_name[0:4]
        if len(res_name) > 3: res_name = res_name[0:3]
        if res_num > 9999: res_num = res_num % 10000
        if len(chain) > 1: chain = chain[0:1]
        return atom_name, res_name, res_num, chain

    @classmethod
    def _make_pdbqt_line(cls, setup, atom_idx, resinfo_set, count):
        """ """
        record_type = "ATOM"
        alt_id = " "
        pdbinfo = setup.pdbinfo[atom_idx]
        if pdbinfo is None:
            pdbinfo = pdbutils.PDBAtomInfo('', '', 0, '')
        resinfo = pdbutils.PDBResInfo(pdbinfo.resName, pdbinfo.resNum, pdbinfo.chain)
        resinfo_set.add(resinfo)
        atom_name, res_name, res_num, chain = cls._get_pdbinfo_fitting_pdb_chars(pdbinfo)
        in_code = ""
        occupancy = 1.0
        temp_factor = 0.0
        coord = setup.coord[atom_idx]
        atom_type = setup.get_atom_type(atom_idx)
        charge = setup.charge[atom_idx]
        atom = "{:6s}{:5d} {:^4s}{:1s}{:3s} {:1s}{:4d}{:1s}   {:8.3f}{:8.3f}{:8.3f}{:6.2f}{:6.2f}    {:6.3f} {:<2s}"

        pdbqt_line = atom.format(record_type, count, pdbinfo.name, alt_id, res_name, chain,
                           res_num, in_code, float(coord[0]), float(coord[1]), float(coord[2]),
                           occupancy, temp_factor, charge, atom_type)
        return pdbqt_line, resinfo_set

    @classmethod
    def _walk_graph_recursive(cls, setup, node, data, edge_start=0, first=False):
        """ recursive walk of rigid bodies"""
        
        if first:
            data["pdbqt_buffer"].append('ROOT')
            member_pool = sorted(setup.flexibility_model['rigid_body_members'][node])
        else:
            member_pool = setup.flexibility_model['rigid_body_members'][node][:]
            member_pool.remove(edge_start)
            member_pool = [edge_start] + member_pool

        for member in member_pool:
            if setup.atom_ignore[member] == 1:
                continue
            pdbqt_line, resinfo_set = cls._make_pdbqt_line(setup, member, data["resinfo_set"], data["count"])
            data["resinfo_set"] = resinfo_set # written as if _make_pdbqt_line() doesn't modify its args (for readability)
            data["pdbqt_buffer"].append(pdbqt_line)
            data["numbering"][member] = data["count"] # count starts at 1
            data["count"] += 1

        if first:
            data["pdbqt_buffer"].append('ENDROOT')

        data["visited"].append(node)

        for neigh in setup.flexibility_model['rigid_body_graph'][node]:
            if neigh in data["visited"]:
                continue

            # Write the branch
            begin, next_index = setup.flexibility_model['rigid_body_connectivity'][node, neigh]

            # do not write branch (or anything downstream) if any of the two atoms
            # defining the rotatable bond are ignored
            if setup.atom_ignore[begin] or setup.atom_ignore[next_index]:
                continue

            begin = data["numbering"][begin]
            end = data["count"]

            data["pdbqt_buffer"].append("BRANCH %3d %3d" % (begin, end))
            data = cls._walk_graph_recursive(setup, neigh, data, edge_start=next_index)
            data["pdbqt_buffer"].append("ENDBRANCH %3d %3d" % (begin, end))
        
        return data

    @classmethod
    def write_string(cls, setup, add_index_map=False, remove_smiles=False, bad_charge_ok=False):
        """Output a PDBQT file as a string.

        Args:
            setup: MoleculeSetup

        Returns:
            str:  PDBQT string of the molecule
            bool: success
            str:  error message
        """

        success = True
        error_msg = ""
        
        if setup.has_implicit_hydrogens():
            error_msg += "molecule has implicit hydrogens (name=%s)\n" % setup.get_mol_name()
            success = False

        for idx, atom_type in setup.atom_type.items():
            if setup.atom_ignore[idx]:
                continue
            if atom_type is None:
                error_msg += 'atom number %d has None type, mol name: %s\n' % (idx, setup.get_mol_name())
                success = False
            c = setup.charge[idx]
            if not bad_charge_ok and (type(c) != float and type(c) != int or math.isnan(c) or math.isinf(c)):
                error_msg += 'atom number %d has non finite charge, mol name: %s, charge: %s\n' % (idx, setup.get_mol_name(), str(c))
                success = False

        if not success:
            pdbqt_string = ""
            return pdbqt_string, success, error_msg

        data = {
            "visited": [],
            "numbering": {},
            "pdbqt_buffer": [],
            "count": 1,
            "resinfo_set": set(),
        }
        atom_counter = {}

        torsdof = len(setup.flexibility_model['rigid_body_graph']) - 1

        if 'torsions_org' in setup.flexibility_model:
            torsdof_org = setup.flexibility_model['torsions_org']
            data["pdbqt_buffer"].append('REMARK Flexibility Score: %2.2f' % setup.flexibility_model['score'] )
            active_tors = torsdof_org
        else:
            active_tors = torsdof

        data = cls._walk_graph_recursive(setup, setup.flexibility_model["root"], data, first=True)

        if add_index_map:
            for i, remark_line in enumerate(cls.remark_index_map(setup, data["numbering"])):
                # Need to use 'insert' because data["numbering"]
                # is populated in self._walk_graph_recursive.
                data["pdbqt_buffer"].insert(i, remark_line)

        if not remove_smiles:
            smiles, order = setup.get_smiles_and_order()
            missing_h = [] # hydrogens which are not in the smiles
            strings_h_parent = []
            for key in data["numbering"]:
                if key in setup.atom_pseudo: continue
                if key not in order:
                    if setup.get_element(key) != 1:
                        error_msg += "non-Hydrogen atom unexpectedely missing from smiles!?"
                        error_msg += " (mol name: %s)\n" % setup.get_mol_name()
                        pdbqt_string = ""
                        success = False
                        return pdbqt_string, success, error_msg
                    missing_h.append(key)
                    parents = setup.get_neigh(key)
                    parents = [i for i in parents if i < setup.atom_true_count] # exclude pseudos
                    if len(parents) != 1:
                        error_msg += "expected hydrogen to be bonded to exactly one atom"
                        error_msg += " (mol name: %s)\n" % setup.get_mol_name()
                        pdbqt_string = ""
                        success = False
                        return pdbqt_string, success, error_msg
                    parent_idx = order[parents[0]] # already 1-indexed
                    string = ' %d %d' % (parent_idx, data["numbering"][key]) # key 0-indexed; _numbering[key] 1-indexed
                    strings_h_parent.append(string)
            remarks_h_parent = cls.break_long_remark_lines(strings_h_parent, "REMARK H PARENT")
            remark_prefix = "REMARK SMILES IDX"
            remark_idxmap = cls.remark_index_map(setup, data["numbering"], order, remark_prefix, missing_h)
            remarks = []
            remarks.append("REMARK SMILES %s" % smiles) # break line at 79 chars?
            remarks.extend(remark_idxmap)
            remarks.extend(remarks_h_parent)

            for i, remark_line in enumerate(remarks):
                # Need to use 'insert' because data["numbering"]
                # is populated in self._walk_graph_recursive.
                data["pdbqt_buffer"].insert(i, remark_line)

        if False: #self.setup.is_protein_sidechain:
            if len(data["resinfo_set"]) > 1:
                print("Warning: more than a single resName, resNum, chain in flexres", file=sys.stderr)
                print(data["resinfo_set"], file=sys.stderr)
            resinfo = list(data["resinfo_set"])[0]
            pdbinfo = pdbutils.PDBAtomInfo('', resinfo.resName, resinfo.resNum, resinfo.chain)
            _, res_name, res_num, chain = cls._get_pdbinfo_fitting_pdb_chars(pdbinfo)
            resinfo_string = "{:3s} {:1s}{:4d}".format(res_name, chain, res_num)
            data["pdbqt_buffer"].insert(0, 'BEGIN_RES %s' % resinfo_string)
            data["pdbqt_buffer"].append('END_RES %s' % resinfo_string)
        else: # no TORSDOF in flexres
            # torsdof is always going to be the one of the rigid, non-macrocyclic one
            data["pdbqt_buffer"].append('TORSDOF %d' % active_tors)

        pdbqt_string =  '\n'.join(data["pdbqt_buffer"]) + '\n'
        return pdbqt_string, success, error_msg

    @classmethod
    def remark_index_map(cls, setup, numbering, order=None, prefix="REMARK INDEX MAP", missing_h=()):
        """ write mapping of atom indices from input molecule to output PDBQT
            order[ob_index(i.e. 'key')] = smiles_index
        """

        if order is None: order = {key: key+1 for key in numbering} # key+1 breaks OB
        #max_line_length = 79
        #remark_lines = []
        #line = prefix
        strings = []
        for key in numbering:
            if key in setup.atom_pseudo: continue
            if key in missing_h: continue
            string = " %d %d" % (order[key], numbering[key])
            strings.append(string)
        return cls.break_long_remark_lines(strings, prefix)
        #    candidate_text = " %d %d" % (order[key], self._numbering[key])
        #    if (len(line) + len(candidate_text)) < max_line_length:
        #        line += candidate_text
        #    else:
        #        remark_lines.append(line)
        #        line = 'REMARK INDEX MAP' + candidate_text
        #remark_lines.append(line)
        #return remark_lines

    @staticmethod
    def break_long_remark_lines(strings, prefix, max_line_length=79):
        remarks = [prefix]
        for string in strings:
            if (len(remarks[-1]) + len(string)) < max_line_length:
                remarks[-1] += string
            else:
                remarks.append(prefix + string)
        return remarks

    @staticmethod
    def adapt_pdbqt_for_autodock4_flexres(pdbqt_string, res, chain, num):
        """ adapt pdbqt_string to be compatible with AutoDock4 requirements:
             - first and second atoms named CA and CB
             - write BEGIN_RES / END_RES
             - remove TORSDOF
            this is for covalent docking (tethered)
        """
        new_string = "BEGIN_RES %s %s %s\n" % (res, chain, num)
        atom_number = 0
        for line in pdbqt_string.split("\n"):
            if line == "":
                continue
            if line.startswith("TORSDOF"):
                continue
            if line.startswith("ATOM"):
                atom_number+=1
                if atom_number == 1:
                    line = line[:13] + 'CA' + line[15:]
                elif atom_number == 2:
                    line = line[:13] + 'CB' + line[15:]
                new_string += line + '\n'
                continue
            new_string += line + '\n'
        new_string += "END_RES %s %s %s\n" % (res, chain, num)
        return new_string
